"""This module contains the ConvexHullLearner class."""
import logging
import os
from pathlib import Path
from typing import Dict, List, Tuple, Optional

import matplotlib.pyplot as plt
import numpy as np
from pandas import DataFrame, Series
from pddl_plus_parser.models import Precondition, PDDLFunction
from scipy.spatial import ConvexHull, convex_hull_plot_2d, QhullError

from sam_learning.core import NotSafeActionError
from sam_learning.core.learning_types import ConditionType, EquationSolutionType
from sam_learning.core.numeric_learning.numeric_utils import (
    construct_multiplication_strings,
    construct_linear_equation_string,
    prettify_coefficients,
    construct_numeric_conditions,
    filter_constant_features,
    detect_linear_dependent_features,
)

EPSILON = 1e-10


class NaiveConvexHullLearner:
    """Class that learns the convex hull of the preconditions of an action."""

    logger: logging.Logger
    action_name: str
    domain_functions: Dict[str, PDDLFunction]
    convex_hull_error_file_path: Path

    def __init__(self, action_name: str, domain_functions: Dict[str, PDDLFunction]):
        self.logger = logging.getLogger(__name__)
        self.convex_hull_error_file_path = Path(os.environ["CONVEX_HULL_ERROR_PATH"])
        self.action_name = action_name
        self.domain_functions = domain_functions

    def _execute_convex_hull(self, points: np.ndarray, display_mode: bool = True) -> Tuple[List[List[float]], List[float]]:
        """Runs the convex hull algorithm on the given input points.

        :param points: the points to run the convex hull algorithm on.
        :param display_mode: whether to display the convex hull.
        :return: the coefficients of the planes that represent the convex hull and the border point.
        """
        hull = ConvexHull(points)
        self._display_convex_hull(display_mode, hull, points.shape[1])

        A = hull.equations[:, : points.shape[1]]
        b = -hull.equations[:, points.shape[1]]
        coefficients = [prettify_coefficients(row) for row in A]
        border_point = prettify_coefficients(b)
        return coefficients, border_point

    def _create_convex_hull_linear_inequalities(
        self, points_df: DataFrame, display_mode: bool = True
    ) -> Tuple[List[List[float]], List[float], List[str], Optional[List[str]]]:
        """Create the convex hull and returns the matrix representing the inequalities.

        :param points_df: the dataframe containing the points that represent the values of the function in the states
            of the observations prior to the action's execution.
        :return: the matrix representing the inequalities of the planes created by the convex hull as well as the
            names of the features that are part of the convex hull.

        Note: the returned values represents the linear inequalities of the convex hull, i.e.,  Ax <= b.
        """
        self.logger.debug(f"Creating convex hull for action {self.action_name}.")
        extra_conditions = []
        columns_to_use = points_df.columns.tolist()
        if points_df.shape[1] == 1:
            points = points_df.to_numpy()
            self.logger.debug("The convex hull is single dimensional, creating min-max conditions on the new basis.")
            coefficients = [[-1], [1]]
            border_point = prettify_coefficients([-points.min(), points.max()])

        else:
            no_constant_columns_matrix, equality_strs, removed_fluents = filter_constant_features(points_df)
            filtered_previous_state_matrix, column_equality_strs, removed_fluents = detect_linear_dependent_features(no_constant_columns_matrix)
            coefficients, border_point = self._execute_convex_hull(filtered_previous_state_matrix.to_numpy(), display_mode)
            extra_conditions = [*equality_strs, *column_equality_strs]
            columns_to_use = filtered_previous_state_matrix.columns.tolist()

        return coefficients, border_point, columns_to_use, extra_conditions

    @staticmethod
    def _construct_pddl_inequality_scheme(
        coefficient_matrix: np.ndarray, border_points: np.ndarray, headers: List[str], sign_to_use: str = "<="
    ) -> List[str]:
        """Construct the inequality strings in the appropriate PDDL format.

        :param coefficient_matrix: the matrix containing the coefficient vectors for each inequality.
        :param border_points: the convex hull point which ensures that Ax <= b.
        :param headers: the headers of the columns in the coefficient matrix.
        :param sign_to_use: the sign to use in the inequalities (for cases when we want to use equalities).
        :return: the inequalities PDDL formatted strings.
        """
        inequalities = set()
        for inequality_coefficients, border_point in zip(coefficient_matrix, border_points):
            multiplication_functions = construct_multiplication_strings(inequality_coefficients, headers)
            constructed_left_side = construct_linear_equation_string(multiplication_functions)
            inequalities.add(f"({sign_to_use} {constructed_left_side} {border_point})")

        return list(inequalities)

    def _display_convex_hull(self, display_mode: bool, hull: ConvexHull, num_dimensions: int) -> None:
        """Displays the convex hull in as a plot.

        :param display_mode: whether to display the plot.
        :param hull: the convex hull to display.
        :param num_dimensions: the number of dimensions of the original data.
        """
        if not display_mode:
            return

        if num_dimensions == 2:
            _ = convex_hull_plot_2d(hull)
            plt.title(f"{self.action_name} - convex hull")
            plt.show()

        elif num_dimensions == 3:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            ax.plot_trisurf(hull.points[:, 0], hull.points[:, 1], hull.points[:, 2], triangles=hull.simplices)
            plt.title(f"{self.action_name} - convex hull")
            plt.show()

    def _construct_single_dimension_inequalities(self, single_column_df: Series, equality_strs: List[str] = []) -> Precondition:
        """Construct a single dimension precondition representation.

        :param single_column_df: the fluent only fluent that is relevant to the preconditions' creation.
        :param equality_strs: the equality conditions that are already present in the preconditions.
        :return: the preconditions string and the condition type.
        """
        min_value = single_column_df.min()
        max_value = single_column_df.max()
        relevant_fluent = single_column_df.name
        if min_value == max_value:
            conditions = [f"(= {relevant_fluent} {min_value})"]
        else:
            conditions = [f"(>= {relevant_fluent} {min_value})", f"(<= {relevant_fluent} {max_value})"]

        conditions.extend(equality_strs)
        return construct_numeric_conditions(conditions, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions)

    def construct_safe_linear_inequalities(self, state_storge: Dict[str, List[float]], relevant_fluents: Optional[List[str]] = []) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :return: the inequality strings and the type of equations that were constructed (injunctive / disjunctive)
        """
        irrelevant_fluents = [fluent for fluent in state_storge.keys() if fluent not in relevant_fluents] if relevant_fluents is not None else []
        state_data = DataFrame(state_storge).drop_duplicates().drop(columns=irrelevant_fluents)
        if relevant_fluents is not None and len(relevant_fluents) == 1:
            self.logger.debug("Only one dimension is needed in the preconditions!")
            return self._construct_single_dimension_inequalities(state_data.loc[:, relevant_fluents[0]])

        try:
            A, b, column_names, additional_projection_conditions = self._create_convex_hull_linear_inequalities(state_data, display_mode=False)
            inequalities_strs = self._construct_pddl_inequality_scheme(A, b, column_names)
            if additional_projection_conditions is not None:
                inequalities_strs.extend(additional_projection_conditions)

            return construct_numeric_conditions(inequalities_strs, condition_type=ConditionType.conjunctive, domain_functions=self.domain_functions)

        except (QhullError, ValueError):
            self.logger.warning(
                "Convex hull failed to create a convex hull, using disjunctive preconditions "
                "(probably since the rank of the matrix is 2 and it cannot create a degraded "
                "convex hull)."
            )
            raise NotSafeActionError(
                name=self.action_name, reason="Convex hull failed to create a convex hull", solution_type=EquationSolutionType.convex_hull_not_found
            )



"""module tests for the Numeric SAM learning algorithm"""
import numpy
import pytest
from pddl_plus_parser.lisp_parsers import ProblemParser, TrajectoryParser, DomainParser
from pddl_plus_parser.models import Domain, Problem, Observation, Predicate
from pytest import fixture

from sam_learning.learners.incremental_numeric_sam import IncrementalNumericSAMLearner
from tests.consts import (
    SATELLITE_PROBLEMATIC_PROBLEM_PATH,
    SATELLITE_PROBLEMATIC_NUMERIC_TRAJECTORY_PATH,
    sync_snapshot,
    MINECRAFT_MEDIUM_DOMAIN_PATH,
    MINECRAFT_MEDIUM_TRAJECTORY_PATH,
    MINECRAFT_SMALL_DOMAIN_PATH,
    MINECRAFT_SMALL_TRAJECTORY_PATH,
    COUNTERS_POLYNOMIAL_DOMAIN_PATH,
    COUNTERS_POLYNOMIAL_PROBLEMS_PATH,
    EXAMPLES_DIR_PATH,
    FARMLAND_DOMAIN_PATH,
    FARMLAND_TRAJECTORIES_DIRECTORY,
    SAILING_LEARNED_DOMAIN_PATH,
    SAILING_TRAJECTORIES_DIRECTORY,
    DRIVERLOG_POLY_TRAJECTORY_PATH,
    DRIVERLOG_POLY_DOMAIN_PATH,
    DRIVERLOG_POLY_PROBLEM_PATH,
)


@fixture()
def driverlog_polynomial_domain() -> Domain:
    return DomainParser(DRIVERLOG_POLY_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def driverlog_polynomial_problem(driverlog_polynomial_domain: Domain) -> Problem:
    return ProblemParser(problem_path=DRIVERLOG_POLY_PROBLEM_PATH, domain=driverlog_polynomial_domain).parse_problem()


@fixture()
def driverlog_polynomial_observation(driverlog_polynomial_domain: Domain,
                                     driverlog_polynomial_problem: Problem) -> Observation:
    return TrajectoryParser(driverlog_polynomial_domain, driverlog_polynomial_problem).parse_trajectory(
        DRIVERLOG_POLY_TRAJECTORY_PATH)


@fixture()
def driverlog_polynomial_nsam(driverlog_polynomial_domain: Domain) -> IncrementalNumericSAMLearner:
    return IncrementalNumericSAMLearner(driverlog_polynomial_domain, polynomial_degree=0, qhull_options="Q12")


@fixture()
def satellite_problem_problematic(satellite_numeric_domain: Domain) -> Problem:
    return ProblemParser(problem_path=SATELLITE_PROBLEMATIC_PROBLEM_PATH,
                         domain=satellite_numeric_domain).parse_problem()


@fixture()
def satellite_observation_problematic(satellite_numeric_domain: Domain,
                                      satellite_problem_problematic: Problem) -> Observation:
    return TrajectoryParser(satellite_numeric_domain, satellite_problem_problematic).parse_trajectory(
        SATELLITE_PROBLEMATIC_NUMERIC_TRAJECTORY_PATH)


@fixture()
def minecraft_medium_domain() -> Domain:
    return DomainParser(MINECRAFT_MEDIUM_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def farmland_domain() -> Domain:
    return DomainParser(FARMLAND_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def sailing_domain() -> Domain:
    return DomainParser(SAILING_LEARNED_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def minecraft_medium_observation(minecraft_medium_domain: Domain) -> Observation:
    return TrajectoryParser(minecraft_medium_domain).parse_trajectory(MINECRAFT_MEDIUM_TRAJECTORY_PATH)


@fixture()
def minecraft_medium_sam(minecraft_medium_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(minecraft_medium_domain, qhull_options="Q12")
    return nsam


@fixture()
def minecraft_small_domain() -> Domain:
    return DomainParser(MINECRAFT_SMALL_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def counters_poly_domain() -> Domain:
    return DomainParser(COUNTERS_POLYNOMIAL_DOMAIN_PATH, partial_parsing=True).parse_domain()


@fixture()
def minecraft_small_observation(minecraft_small_domain: Domain) -> Observation:
    return TrajectoryParser(minecraft_small_domain).parse_trajectory(MINECRAFT_SMALL_TRAJECTORY_PATH)


@fixture()
def minecraft_small_sam(minecraft_small_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(minecraft_small_domain)
    return nsam


@fixture()
def depot_nsam(depot_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(depot_domain)
    return nsam


@fixture()
def satellite_nsam(satellite_numeric_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(satellite_numeric_domain)
    return nsam


@fixture()
def minecraft_nsam(minecraft_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(minecraft_domain)
    return nsam


@fixture()
def counters_poly_nsam(counters_poly_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(counters_poly_domain, polynomial_degree=1)
    return nsam


@fixture()
def farmland_nsam(farmland_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(farmland_domain, polynomial_degree=0, qhull_options="Q12")
    return nsam


@fixture()
def sailing_nsam(sailing_domain: Domain) -> IncrementalNumericSAMLearner:
    nsam = IncrementalNumericSAMLearner(sailing_domain, polynomial_degree=0)
    return nsam


def test_add_new_action_adds_sets_the_dataset_with_new_observation_for_the_previous_state(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    assert len(depot_nsam.storage[action_call.name].convex_hull_learner.data) == 1


def test_add_new_action_adds_discrete_preconditions_to_the_learned_action(depot_nsam: IncrementalNumericSAMLearner,
                                                                          depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if
                                     isinstance(precondition, Predicate)]
    assert len(action_discrete_preconditions) > 0


def test_add_new_action_adds_the_required_predicates_to_the_action_preconditions(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if
                                     isinstance(precondition, Predicate)]
    preconditions_str = {precondition.untyped_representation for precondition in action_discrete_preconditions}
    assert preconditions_str.issuperset({"(at ?x ?y)"})


def test_update_action_when_adding_a_point_that_was_already_added_to_the_dataset_does_not_add_it_again(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    depot_nsam.update_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    assert len(depot_nsam.storage[action_call.name].convex_hull_learner.data) == 1
    assert depot_nsam.storage[action_call.name].convex_hull_learner._gsp_base is None
    assert depot_nsam.storage[action_call.name].convex_hull_learner._complementary_base is None


def test_update_action_when_adding_a_different_sample_adds_the_data_to_the_gsp_basis(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    test_action_name = "drive"
    relevant_triplet_indexes = [0, 4]
    for triplet in relevant_triplet_indexes:
        pre_state = depot_observation.components[triplet].previous_state
        action_call = depot_observation.components[triplet].grounded_action_call
        next_state = depot_observation.components[triplet].next_state
        sync_snapshot(depot_nsam, depot_observation.components[triplet], depot_observation.grounded_objects)
        if triplet == 0:
            depot_nsam.add_new_action(grounded_action=action_call, previous_state=pre_state, next_state=next_state)
        else:
            depot_nsam.update_action(grounded_action=action_call, previous_state=pre_state, next_state=next_state)

    assert len(depot_nsam.storage[test_action_name].convex_hull_learner.data) == 2
    assert depot_nsam.storage[test_action_name].convex_hull_learner._gsp_base is not None
    assert depot_nsam.storage[test_action_name].convex_hull_learner._complementary_base is not None
    assert depot_nsam.storage[test_action_name].convex_hull_learner._convex_hull is None


def test_update_action_for_the_second_time_adds_the_point_to_the_dataset_and_creates_initial_convex_hull(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    test_action_name = "drive"
    relevant_triplet_indexes = [0, 4, 5]
    for triplet in relevant_triplet_indexes:
        pre_state = depot_observation.components[triplet].previous_state
        action_call = depot_observation.components[triplet].grounded_action_call
        next_state = depot_observation.components[triplet].next_state
        sync_snapshot(depot_nsam, depot_observation.components[triplet], depot_observation.grounded_objects)
        if triplet == 0:
            depot_nsam.add_new_action(grounded_action=action_call, previous_state=pre_state, next_state=next_state)
        else:
            depot_nsam.update_action(grounded_action=action_call, previous_state=pre_state, next_state=next_state)

    assert len(depot_nsam.storage[test_action_name].convex_hull_learner.data) == 3
    assert depot_nsam.storage[test_action_name].convex_hull_learner._gsp_base is not None
    assert depot_nsam.storage[test_action_name].convex_hull_learner._complementary_base is not None
    assert depot_nsam.storage[test_action_name].convex_hull_learner._convex_hull is not None


def test_update_action_adds_discrete_preconditions_to_the_learned_action(depot_nsam: IncrementalNumericSAMLearner,
                                                                         depot_observation: Observation):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    depot_nsam.update_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if
                                     isinstance(precondition, Predicate)]
    assert len(action_discrete_preconditions) > 0


def test_update_action_adds_the_required_predicates_to_the_action_preconditions(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    initial_state = depot_observation.components[0].previous_state
    action_call = depot_observation.components[0].grounded_action_call
    next_state = depot_observation.components[0].next_state
    sync_snapshot(depot_nsam, depot_observation.components[0], depot_observation.grounded_objects)
    depot_nsam.add_new_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    depot_nsam.update_action(grounded_action=action_call, previous_state=initial_state, next_state=next_state)
    learned_action = depot_nsam.partial_domain.actions[action_call.name]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if
                                     isinstance(precondition, Predicate)]
    preconditions_str = {precondition.untyped_representation for precondition in action_discrete_preconditions}
    assert preconditions_str.issuperset({"(at ?x ?y)"})


def test_handle_single_trajectory_component_does_not_remove_the_required_predicates_to_the_action_preconditions(
        depot_nsam: IncrementalNumericSAMLearner, depot_observation: Observation
):
    depot_nsam.current_trajectory_objects = depot_observation.grounded_objects
    depot_nsam.handle_single_trajectory_component(depot_observation.components[0])
    learned_action = depot_nsam.partial_domain.actions["drive"]
    action_discrete_preconditions = [precondition for _, precondition in learned_action.preconditions if
                                     isinstance(precondition, Predicate)]
    preconditions_str = {precondition.untyped_representation for precondition in action_discrete_preconditions}
    assert preconditions_str.issuperset({"(at ?x ?y)"})


def test_learn_action_model_returns_learned_model(depot_nsam: IncrementalNumericSAMLearner,
                                                  depot_observation: Observation):
    learned_model, learning_metadata = depot_nsam.learn_action_model([depot_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_for_satellite_domain_returns_learned_model(
        satellite_nsam: IncrementalNumericSAMLearner, satellite_numeric_observation: Observation
):
    learned_model, learning_metadata = satellite_nsam.learn_action_model([satellite_numeric_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_for_satellite_with_problematic_trajectory_domain_returns_learned_model(
        satellite_nsam: IncrementalNumericSAMLearner, satellite_observation_problematic: Observation
):
    learned_model, learning_metadata = satellite_nsam.learn_action_model([satellite_observation_problematic])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_minecraft_domain_creates_domain_with_correct_preconditions_and_effects(
        minecraft_nsam: IncrementalNumericSAMLearner, minecraft_observation: Observation
):
    learned_model, learning_metadata = minecraft_nsam.learn_action_model([minecraft_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_minecraft_medium_domain_creates_domain_with_correct_preconditions_and_effects(
        minecraft_medium_sam: IncrementalNumericSAMLearner, minecraft_medium_observation: Observation
):
    learned_model, learning_metadata = minecraft_medium_sam.learn_action_model([minecraft_medium_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_minecraft_small_domain_creates_domain_with_correct_preconditions_and_effects(
        minecraft_small_sam: IncrementalNumericSAMLearner, minecraft_small_observation: Observation
):
    learned_model, learning_metadata = minecraft_small_sam.learn_action_model([minecraft_small_observation])
    print()
    print(learning_metadata)
    print(learned_model.to_pddl())


def test_learn_action_model_with_counters_poly_domain_enables_learning_polynomial_actions_and_does_not_raise_error_while_learning_the_convex_hull(
        counters_poly_nsam: IncrementalNumericSAMLearner, counters_poly_domain: Domain
):
    observations = []
    for problem_path in COUNTERS_POLYNOMIAL_PROBLEMS_PATH:
        trajectory_path = EXAMPLES_DIR_PATH / f"{problem_path.stem}.trajectory"
        problem = ProblemParser(problem_path, counters_poly_domain).parse_problem()
        observation = TrajectoryParser(counters_poly_domain, problem).parse_trajectory(trajectory_path)
        observations.append(observation)

    try:
        learned_model, learning_metadata = counters_poly_nsam.learn_action_model(observations)
        print()
        print(learning_metadata)
        print(learned_model.to_pddl())

    except Exception:
        pytest.fail()


def test_learn_action_model_when_learning_sailing_domain_from_test_dataset_returns_correct_domain_with_unshifted_conditions(
        sailing_nsam: IncrementalNumericSAMLearner, sailing_domain: Domain
):
    observations = []
    for problem_path in SAILING_TRAJECTORIES_DIRECTORY.glob("pfile*.pddl"):
        trajectory_path = SAILING_TRAJECTORIES_DIRECTORY / f"{problem_path.stem}.trajectory"
        problem = ProblemParser(problem_path, sailing_domain).parse_problem()
        observation = TrajectoryParser(sailing_domain, problem).parse_trajectory(trajectory_path)
        observations.append(observation)

    learned_model, learning_metadata = sailing_nsam.learn_action_model(observations)
    for action in learned_model.actions:
        if sailing_nsam.storage[action].convex_hull_learner._spanning_standard_base:
            assert sailing_nsam.storage[action].convex_hull_learner._convex_hull is not None
            assert (
                    sailing_nsam.storage[action].convex_hull_learner._convex_hull.points
                    == sailing_nsam.storage[action].convex_hull_learner.data.to_numpy(dtype=numpy.float32)
            ).all()

    print()
    print(learned_model.to_pddl())


def test_learn_action_model_with_driverlog_domain_does_not_fail_with_error(
        driverlog_polynomial_nsam: IncrementalNumericSAMLearner, driverlog_polynomial_observation: Observation
):
    try:
        learned_model, learning_metadata = driverlog_polynomial_nsam.learn_action_model(
            [driverlog_polynomial_observation])
        print()
        print(learning_metadata)
        print(learned_model.to_pddl())
    except Exception as e:
        assert False, f"An error occurred: {e}"



"""Module that stores amd learns an action's numeric state fluents."""
import logging
from typing import Dict, List, Optional, Set, Tuple

import numpy as np
from pddl_plus_parser.models import PDDLFunction, Precondition, NumericalExpressionTree

from sam_learning.core.numeric_learning.incremental_convex_hull_learner import IncrementalConvexHullLearner
from sam_learning.core.numeric_learning.numeric_fluent_learner_algorithm import NumericFluentStateStorage
from sam_learning.core.numeric_learning.numeric_utils import create_monomials, create_polynomial_string

np.seterr(divide="ignore", invalid="ignore")


class IncrementalNumericFluentStateStorage(NumericFluentStateStorage):
    """Stores and learned the numeric state fluents of a single action."""

    def __init__(self, action_name: str, domain_functions: Dict[str, PDDLFunction], polynom_degree: int = 0):
        super().__init__(action_name, domain_functions)
        self.logger = logging.getLogger(__name__)
        self.monomials = create_monomials(list(domain_functions.keys()), polynom_degree)
        self.convex_hull_learner = IncrementalConvexHullLearner(action_name, domain_functions, polynom_degree)

    def add_to_previous_state_storage(self, state_fluents: Dict[str, PDDLFunction]) -> None:
        """Adds the matched lifted state fluents to the previous state storage.

        :param state_fluents: the lifted state fluents that were matched for the action.
        """
        sample_dataset = {}
        for monomial in self.monomials:
            sample_dataset[create_polynomial_string(monomial)] = np.prod([state_fluents[fluent].value for fluent in monomial])

        self.convex_hull_learner.add_new_point(sample_dataset)
        self.linear_regression_learner.add_new_observation(sample_dataset, store_in_prev_state=True)

    def construct_safe_linear_inequalities(self, relevant_fluents: Optional[List[str]] = None) -> Precondition:
        """Constructs the linear inequalities strings that will be used in the learned model later.

        :return: The precondition that contains the linear inequalities.
        """
        self.logger.info("Constructing the safe linear inequalities.")
        return self.convex_hull_learner.construct_convex_hull_inequalities()

    def construct_assignment_equations(self) -> Tuple[Set[NumericalExpressionTree], Optional[Precondition], bool]:
        """Constructs the assignment statements for the action according to the changed value functions.

        :return: the constructed assignment statements.
        """
        self.logger.info("Constructing the assignment equations.")
        return self.linear_regression_learner.construct_assignment_equations()



from .util_types import LearningAlgorithmType, SolverType, SolutionOutputTypes
from .util_types import NegativePreconditionPolicy
from .k_fold_split import KFoldSplit
from .distributed_k_fold_split import DistributedKFoldSplit



"""Test the extended functionality of the predicate matcher."""

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import GroundedPredicate, Domain, ActionCall, Predicate, Observation
from pytest import fixture

from sam_learning.core import PredicatesMatcher
from tests.consts import (
    LOCATION_TYPE,
    AGENT_TYPE,
    PART_TYPE,
    TREATMENT_STATUS_TYPE,
    SURFACE_TYPE,
    CITY_TYPE,
    OBJECT_TYPE,
    AIRPLANE_TYPE,
    TRUCK_TYPE,
    COLOR_TYPE,
    COUNT_TYPE,
    BARMAN_CLASSICAL_DOMAIN_PATH,
)

TRUCK_AT_LOCATION_GROUNDED_PREDICATE = GroundedPredicate(
    name="at", signature={"?obj": OBJECT_TYPE, "?loc": LOCATION_TYPE}, object_mapping={"?obj": "tru1", "?loc": "pos1"}
)

IN_CITY_GROUNDED_PREDICATE = GroundedPredicate(
    name="in-city",
    signature={"?agent": TRUCK_TYPE, "?loc": LOCATION_TYPE, "?city": CITY_TYPE},
    object_mapping={"?agent": "tru1", "?loc": "pos1", "?city": "city1"},
)

IN_GROUNDED_PREDICATE = GroundedPredicate(
    name="in",
    signature={"?obj": OBJECT_TYPE, "?airplane": AIRPLANE_TYPE},
    object_mapping={"?obj": "pack1", "?airplane": "airplane1"},
)

TREATMENT_GROUNDED_PREDICATE = GroundedPredicate(
    name="treatment",
    signature={"?obj": PART_TYPE, "?treatment": TREATMENT_STATUS_TYPE},
    object_mapping={"?obj": "obj1", "?treatment": "untreated"},
)

IS_SMOOTH_GROUNDED_PREDICATE = GroundedPredicate(
    name="is-smooth", signature={"?surface": SURFACE_TYPE}, object_mapping={"?surface": "verysmooth"}
)

NEXT_GROUNDED_PREDICATE = GroundedPredicate(
    name="next", signature={"?n1": COUNT_TYPE, "?n2": COUNT_TYPE}, object_mapping={"?n1": "f1", "?n2": "f1"}
)


@fixture()
def logistics_predicate_matcher(logistics_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=logistics_domain)


@fixture()
def woodworking_predicate_matcher(woodworking_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=woodworking_domain)


@fixture()
def elevators_predicate_matcher(elevators_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=elevators_domain)


@fixture()
def depot_predicate_matcher(depot_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=depot_domain)


@fixture()
def spider_predicate_matcher(spider_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=spider_domain)


@fixture()
def nurikabe_predicate_matcher(nurikabe_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=nurikabe_domain)


@fixture()
def barman_predicate_matcher(barman_domain: Domain) -> PredicatesMatcher:
    return PredicatesMatcher(domain=barman_domain)


def test_match_predicate_to_action_with_no_match_returns_empty_list(logistics_predicate_matcher):
    test_action_call = ActionCall(name="drive-truck", grounded_parameters=["tru2", "pos1", "pos2", "city1"])
    actual_predicates = logistics_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=TRUCK_AT_LOCATION_GROUNDED_PREDICATE, action_call=test_action_call
    )

    assert len(actual_predicates) == 0


def test_match_predicate_to_action_with_no_duplicated_parameters_returns_correct_match(logistics_predicate_matcher):
    test_action_call = ActionCall(name="drive-truck", grounded_parameters=["tru1", "pos1", "pos2", "city1"])
    actual_predicates = logistics_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=TRUCK_AT_LOCATION_GROUNDED_PREDICATE, action_call=test_action_call
    )

    expected_predicate = Predicate(name="at", signature={"?truck": TRUCK_TYPE, "?loc-from": LOCATION_TYPE})

    assert len(actual_predicates) == 1
    assert expected_predicate.untyped_representation == actual_predicates[0].untyped_representation


def test_match_predicate_to_action_with_out_of_order_parameters_returns_correct_match(logistics_predicate_matcher):
    test_action_call = ActionCall(name="unload-airplane", grounded_parameters=["airplane1", "pack1", "pos1"])
    actual_predicates = logistics_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=IN_GROUNDED_PREDICATE, action_call=test_action_call
    )

    expected_predicate = Predicate(name="in", signature={"?obj": OBJECT_TYPE, "?airplane": AIRPLANE_TYPE})

    assert len(actual_predicates) == 1
    assert expected_predicate.untyped_representation == actual_predicates[0].untyped_representation


def test_match_predicate_to_action_with_no_duplicated_parameters_with_consts_returns_correct_match(
    woodworking_predicate_matcher,
):
    test_action_call = ActionCall(name="do-spray-varnish", grounded_parameters=["m1", "obj1", "natural", "rough"])
    actual_predicates = woodworking_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=TREATMENT_GROUNDED_PREDICATE, action_call=test_action_call
    )

    expected_predicate = Predicate(name="treatment", signature={"?x": AGENT_TYPE, "untreated": LOCATION_TYPE})

    assert len(actual_predicates) == 1
    assert expected_predicate.untyped_representation == actual_predicates[0].untyped_representation


def test_match_predicate_to_action_with_two_possible_options_for_match_one_with_constant_and_one_without_returns_two_options(
    woodworking_predicate_matcher,
):
    test_action_call = ActionCall(name="do-immersion-varnish", grounded_parameters=["m1", "obj1", "natural", "rough"])
    test_conflicting_predicate = GroundedPredicate(
        name="colour",
        signature={"?obj": PART_TYPE, "?color": COLOR_TYPE},
        object_mapping={"?obj": "obj1", "?color": "natural"},
    )
    actual_predicates = woodworking_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=test_conflicting_predicate, action_call=test_action_call
    )
    print([p.untyped_representation for p in actual_predicates])

    expected_predicates = [
        Predicate(name="colour", signature={"?x": PART_TYPE, "?newcolour": COLOR_TYPE}),
        Predicate(name="colour", signature={"?x": PART_TYPE, "natural": COLOR_TYPE}),
    ]

    assert len(actual_predicates) == 2
    assert set([p.untyped_representation for p in actual_predicates]) == set(
        [p.untyped_representation for p in expected_predicates]
    )


def test_match_predicate_to_action_with_no_duplicated_parameters_where_predicate_contains_only_consts_returns_correct_match(
    woodworking_predicate_matcher,
):
    test_action_call = ActionCall(name="do-spray-varnish", grounded_parameters=["m1", "obj1", "natural", "verysmooth"])
    actual_predicates = woodworking_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=IS_SMOOTH_GROUNDED_PREDICATE, action_call=test_action_call
    )

    expected_predicates = [
        Predicate(name="is-smooth", signature={"?surface": SURFACE_TYPE}),
        Predicate(name="is-smooth", signature={"verysmooth": SURFACE_TYPE}),
    ]

    assert len(actual_predicates) == 2
    assert [p.untyped_representation for p in expected_predicates] == [
        p.untyped_representation for p in actual_predicates
    ]


def test_match_predicate_to_action_literals_with_complex_action_returns_only_correct_matches(
    elevators_predicate_matcher: PredicatesMatcher, elevators_domain: Domain
):
    test_action_call = ActionCall(name="move-down-slow", grounded_parameters=["slow2-0", "n17", "n16"])
    next_predicate = elevators_domain.predicates["next"]
    test_grounded_predicate = GroundedPredicate(
        name=next_predicate.name,
        signature=next_predicate.signature,
        object_mapping={
            "?n1": "n16",
            "?n2": "n17",
        },
    )

    actual_predicates = elevators_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=test_grounded_predicate, action_call=test_action_call
    )
    assert len(actual_predicates) == 1
    assert actual_predicates[0].untyped_representation == "(next ?f2 ?f1)"


def test_match_predicate_to_action_literals_with_complex_action_with_duplicates_returns_only_correct_matches(
    elevators_predicate_matcher: PredicatesMatcher, elevators_domain: Domain
):
    test_action_call = ActionCall(name="move-down-slow", grounded_parameters=["slow2-0", "n17", "n17"])
    next_predicate = elevators_domain.predicates["next"]
    test_grounded_predicate = GroundedPredicate(
        name=next_predicate.name,
        signature=next_predicate.signature,
        object_mapping={
            "?n1": "n17",
            "?n2": "n17",
        },
    )

    actual_predicates = elevators_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=test_grounded_predicate, action_call=test_action_call
    )
    assert len(actual_predicates) == 2
    assert [p.untyped_representation for p in actual_predicates] == ["(next ?f1 ?f2)", "(next ?f2 ?f1)"]


def test_match_predicate_to_action_with_duplicated_objects_finds_all_possible_matches(logistics_predicate_matcher):
    test_action_call = ActionCall(name="drive-truck", grounded_parameters=["tru1", "pos1", "pos1", "city1"])
    actual_predicates = logistics_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=TRUCK_AT_LOCATION_GROUNDED_PREDICATE, action_call=test_action_call
    )
    expected_predicates = [
        Predicate(name="at", signature={"?truck": AGENT_TYPE, "?loc-from": LOCATION_TYPE}),
        Predicate(name="at", signature={"?truck": AGENT_TYPE, "?loc-to": LOCATION_TYPE}),
    ]

    assert len(actual_predicates) == 2
    assert [p.untyped_representation for p in expected_predicates] == [
        p.untyped_representation for p in actual_predicates
    ]


def test_get_possible_literal_matches_with_single_predicate_without_duplicates_returns_correct_matches(
    logistics_predicate_matcher,
):
    test_action_call = ActionCall(name="drive-truck", grounded_parameters=["tru1", "pos1", "pos2", "city1"])
    test_state_predicates = [TRUCK_AT_LOCATION_GROUNDED_PREDICATE]
    possible_matches = logistics_predicate_matcher.get_possible_literal_matches(test_action_call, test_state_predicates)

    expected_predicate = Predicate(name="at", signature={"?truck": AGENT_TYPE, "?loc-from": LOCATION_TYPE})

    assert len(possible_matches) == 1
    assert expected_predicate.untyped_representation == possible_matches[0].untyped_representation


def test_get_possible_literal_matches_with_single_predicate_with_duplicates_returns_correct_matches(
    logistics_predicate_matcher,
):
    test_action_call = ActionCall(name="drive-truck", grounded_parameters=["tru1", "pos1", "pos1", "city1"])
    test_state_predicates = [TRUCK_AT_LOCATION_GROUNDED_PREDICATE]
    possible_matches = logistics_predicate_matcher.get_possible_literal_matches(test_action_call, test_state_predicates)

    expected_predicates = [
        Predicate(name="at", signature={"?truck": AGENT_TYPE, "?loc-from": LOCATION_TYPE}),
        Predicate(name="at", signature={"?truck": AGENT_TYPE, "?loc-to": LOCATION_TYPE}),
    ]

    assert len(possible_matches) == 2
    assert [p.untyped_representation for p in expected_predicates] == [
        p.untyped_representation for p in possible_matches
    ]


def test_get_possible_literal_matches_with_two_predicate_returns_correct_matches(logistics_predicate_matcher):
    test_action_call = ActionCall(name="drive-truck", grounded_parameters=["tru1", "pos1", "pos2", "city1"])
    test_state_predicates = [TRUCK_AT_LOCATION_GROUNDED_PREDICATE, IN_CITY_GROUNDED_PREDICATE]
    possible_matches = logistics_predicate_matcher.get_possible_literal_matches(test_action_call, test_state_predicates)

    expected_predicates = [
        Predicate(name="at", signature={"?truck": AGENT_TYPE, "?loc-from": LOCATION_TYPE}),
        Predicate(name="in-city", signature={"?truck": AGENT_TYPE, "?loc-from": LOCATION_TYPE, "?city": CITY_TYPE}),
    ]

    assert len(possible_matches) == 2
    assert [p.untyped_representation for p in expected_predicates] == [
        p.untyped_representation for p in possible_matches
    ]


def test_get_possible_literal_matches_from_actual_trajectory_state(
    elevators_predicate_matcher: PredicatesMatcher, elevators_observation: Observation
):
    observation_component = elevators_observation.components[0]
    test_action_call = observation_component.grounded_action_call

    previous_state_predicates = []
    for predicate_set in observation_component.previous_state.state_predicates.values():
        previous_state_predicates.extend(predicate_set)

    possible_matches = elevators_predicate_matcher.get_possible_literal_matches(
        test_action_call, previous_state_predicates
    )
    for matched_lifted_predicate in possible_matches:
        for parameter in matched_lifted_predicate.signature:
            assert parameter in ["?lift", "?f1", "?f2"]


def test_get_possible_literal_from_actual_state_captures_all_needed_predicates_in_a_numeric_domain(
    depot_predicate_matcher: PredicatesMatcher, depot_observation: Observation
):
    observation_component = depot_observation.components[1]
    test_action_call = observation_component.grounded_action_call
    previous_state_predicates = []
    for predicate_set in observation_component.previous_state.state_predicates.values():
        previous_state_predicates.extend(predicate_set)

    possible_matches = depot_predicate_matcher.get_possible_literal_matches(test_action_call, previous_state_predicates)

    actual_preconditions = {"(at ?x ?p)", "(available ?x)", "(at ?y ?p)", "(on ?y ?z)", "(clear ?y)"}
    possible_lifted_matches = [
        matched_lifted_predicate.untyped_representation for matched_lifted_predicate in possible_matches
    ]
    print(possible_lifted_matches)
    assert actual_preconditions.issubset(possible_lifted_matches)


def test_match_predicate_to_action_literals_when_action_and_predicates_have_no_parameters_match_them_to_each_other(
    spider_predicate_matcher: PredicatesMatcher, spider_domain: Domain
):
    test_action_call = ActionCall(name="start-dealing", grounded_parameters=[])
    test_predicate = spider_domain.predicates["currently-dealing"]
    test_state_predicates = [GroundedPredicate(name=test_predicate.name, signature={}, object_mapping={})]
    possible_matches = spider_predicate_matcher.get_possible_literal_matches(test_action_call, test_state_predicates)
    assert len(possible_matches) == 1


def test_match_predicate_to_action_literals_when_given_extra_object_to_match_creates_lifted_matches_correctly(
    nurikabe_predicate_matcher: PredicatesMatcher, nurikabe_domain: Domain
):
    test_action_call = ActionCall(name="move-painting", grounded_parameters=["pos-5-0", "pos-4-0", "g0", "n1", "n0"])
    test_predicate = nurikabe_domain.predicates["part-of"]
    test_state_predicate = GroundedPredicate(
        name=test_predicate.name, signature=test_predicate.signature, object_mapping={"?x": "pos-5-1", "?y": "g0"}
    )
    possible_matches = nurikabe_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=test_state_predicate,
        action_call=test_action_call,
        extra_grounded_object="pos-5-1",
        extra_lifted_object="?cadj",
    )
    assert len(possible_matches) == 1
    assert possible_matches[0].untyped_representation == "(part-of ?cadj ?g)"


def test_match_predicate_to_action_literals_when_given_extra_object_to_match_does_not_match_when_the_parameters_are_wrong(
    nurikabe_predicate_matcher: PredicatesMatcher, nurikabe_domain: Domain
):
    test_action_call = ActionCall(name="move-painting", grounded_parameters=["pos-5-0", "pos-4-0", "g0", "n1", "n0"])
    test_predicate = nurikabe_domain.predicates["part-of"]
    test_state_predicate = GroundedPredicate(
        name=test_predicate.name, signature=test_predicate.signature, object_mapping={"?x": "pos-5-1", "?y": "g0"}
    )
    possible_matches = nurikabe_predicate_matcher.match_predicate_to_action_literals(
        grounded_predicate=test_state_predicate,
        action_call=test_action_call,
        extra_grounded_object="pos-5-2",
        extra_lifted_object="?cadj",
    )
    assert len(possible_matches) == 0


def test_get_possible_literal_matches_with_extra_literal_extends_the_possible_set_of_matches(
    nurikabe_predicate_matcher: PredicatesMatcher, nurikabe_domain: Domain
):
    test_action_call = ActionCall(name="move-painting", grounded_parameters=["pos-5-0", "pos-4-0", "g0", "n1", "n0"])
    test_predicate = nurikabe_domain.predicates["part-of"]
    test_state_predicates = [
        GroundedPredicate(
            name=test_predicate.name, signature=test_predicate.signature, object_mapping={"?x": "pos-5-0", "?y": "g0"}
        ),
        GroundedPredicate(
            name=test_predicate.name, signature=test_predicate.signature, object_mapping={"?x": "pos-5-2", "?y": "g0"}
        ),
    ]
    possible_matches = nurikabe_predicate_matcher.get_possible_literal_matches(
        test_action_call, test_state_predicates, extra_grounded_object="pos-5-2", extra_lifted_object="?cadj"
    )
    assert len(possible_matches) == len(test_state_predicates)


def test_get_possible_literal_matches_with_extra_literal_extends_the_possible_set_of_matches_with_correct_data(
    nurikabe_predicate_matcher: PredicatesMatcher, nurikabe_domain: Domain
):
    test_action_call = ActionCall(name="move", grounded_parameters=["pos-0-0", "pos-0-1"])
    test_predicate = nurikabe_domain.predicates["connected"]
    test_state_predicates = [
        GroundedPredicate(
            name=test_predicate.name,
            signature=test_predicate.signature,
            object_mapping={"?c": "pos-0-2", "?c2": "pos-0-1"},
        ),
        GroundedPredicate(
            name=test_predicate.name,
            signature=test_predicate.signature,
            object_mapping={"?c": "pos-0-2", "?c2": "pos-0-3"},
        ),
        GroundedPredicate(
            name=test_predicate.name,
            signature=test_predicate.signature,
            object_mapping={"?c": "pos-0-2", "?c2": "pos-2-1"},
        ),
    ]
    possible_matches = nurikabe_predicate_matcher.get_possible_literal_matches(
        test_action_call, test_state_predicates, extra_grounded_object="pos-0-2", extra_lifted_object="?cadj"
    )
    assert len(possible_matches) == 1
    assert possible_matches[0].untyped_representation == "(connected ?cadj ?to)"


def test_get_injective_match_returns_a_single_match_with_a_predicate_that_has_is_positive_property_set_to_false_when_given_a_negative_grounded_predicate_as_input(
    logistics_predicate_matcher,
):
    IN_GROUNDED_PREDICATE.is_positive = False
    test_action_call = ActionCall(name="unload-airplane", grounded_parameters=["airplane1", "pack1", "pos1"])
    possible_matches = logistics_predicate_matcher.get_injective_match(IN_GROUNDED_PREDICATE, test_action_call)
    assert possible_matches is not None
    assert possible_matches.is_positive is False


def test_get_injective_match_returns_none_if_there_are_multiple_matches(elevators_predicate_matcher: PredicatesMatcher):
    NEXT_GROUNDED_PREDICATE.is_positive = False
    test_action_call = ActionCall(name="move-up-slow", grounded_parameters=["slow-lift", "f1", "f1"])
    possible_matches = elevators_predicate_matcher.get_injective_match(NEXT_GROUNDED_PREDICATE, test_action_call)
    assert possible_matches is None



"""Module test for the numeric state storage."""
import random

from pandas import DataFrame
from pddl_plus_parser.models import NumericalExpressionTree
from pytest import fixture

from sam_learning.core import NumericFluentStateStorage, construct_non_circular_assignment
from sam_learning.core.numeric_learning.numeric_utils import detect_linear_dependent_features, filter_constant_features
from tests.consts import FUEL_COST_FUNCTION, LOAD_LIMIT_TRAJECTORY_FUNCTION, CURRENT_LOAD_TRAJECTORY_FUNCTION

LOAD_ACTION = "load"

TEST_DOMAIN_FUNCTIONS = {
    "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
    "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
    "(fuel-cost )": FUEL_COST_FUNCTION,
}

random.seed(0)


@fixture()
def load_action_state_fluent_storage() -> NumericFluentStateStorage:
    return NumericFluentStateStorage(action_name=LOAD_ACTION, domain_functions=TEST_DOMAIN_FUNCTIONS)


def test_add_to_previous_state_storage_can_add_single_item_to_the_storage(load_action_state_fluent_storage: NumericFluentStateStorage,):
    LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(411.0)
    CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(121.0)
    FUEL_COST_FUNCTION.set_value(34.0)
    simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
    }
    load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)
    assert len(load_action_state_fluent_storage.convex_hull_learner.data) == 1
    assert load_action_state_fluent_storage.convex_hull_learner.data.iloc[0]["(fuel-cost )"] == [34.0]
    assert load_action_state_fluent_storage.convex_hull_learner.data.iloc[0]["(load_limit ?z)"] == [411.0]
    assert load_action_state_fluent_storage.convex_hull_learner.data.iloc[0]["(current_load ?z)"] == [121.0]


def test_add_to_next_state_storage_can_add_single_item_to_the_storage(load_action_state_fluent_storage: NumericFluentStateStorage,):
    LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(411.0)
    CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(121.0)
    FUEL_COST_FUNCTION.set_value(34.0)
    simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
    }
    load_action_state_fluent_storage.add_to_next_state_storage(simple_state_fluents)
    assert load_action_state_fluent_storage.linear_regression_learner.next_state_data.iloc[0]["(fuel-cost )"] == [34.0]
    assert load_action_state_fluent_storage.linear_regression_learner.next_state_data.iloc[0]["(load_limit ?z)"] == [411.0]
    assert load_action_state_fluent_storage.linear_regression_learner.next_state_data.iloc[0]["(current_load ?z)"] == [121.0]


def test_add_to_previous_state_storage_can_add_multiple_state_values_correctly(load_action_state_fluent_storage: NumericFluentStateStorage,):
    for i in range(10):
        FUEL_COST_FUNCTION.set_value(i)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(i + 1)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(i + 2)
        simple_state_fluents = {
            "(fuel-cost )": FUEL_COST_FUNCTION,
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)

    assert len(load_action_state_fluent_storage.convex_hull_learner.data) == 10


def test_add_to_next_state_storage_can_add_multiple_state_values_correctly(load_action_state_fluent_storage: NumericFluentStateStorage,):
    for i in range(10):
        FUEL_COST_FUNCTION.set_value(i)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(i + 1)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(i + 2)
        simple_state_fluents = {
            "(fuel-cost )": FUEL_COST_FUNCTION,
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_next_state_storage(simple_state_fluents)

    assert len(load_action_state_fluent_storage.linear_regression_learner.next_state_data) == 10
    assert len(load_action_state_fluent_storage.linear_regression_learner.next_state_data.columns.tolist()) == 3


def test_construct_non_circular_assignment_constructs_correct_equation_with_correct_coefficient_sign_on_increase(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    lifted_function = "(current_load ?z)"
    coefficient_map = {"(current_load ?z)": 1.0, "(weight ?y)": 1.0}
    previous_value = 0.0
    next_value = 1.0
    increase_statement = construct_non_circular_assignment(lifted_function, coefficient_map, previous_value, next_value)
    assert increase_statement == "(increase (current_load ?z) (weight ?y))"


def test_construct_non_circular_assignment_constructs_correct_equation_with_correct_coefficient_sign_on_decrease(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    lifted_function = "(current_load ?z)"
    coefficient_map = {"(current_load ?z)": 1.0, "(weight ?y)": -1.0, "(dummy)": 0.0}
    previous_value = 1.0
    next_value = 0.0
    increase_statement = construct_non_circular_assignment(lifted_function, coefficient_map, previous_value, next_value)
    assert increase_statement == "(decrease (current_load ?z) (weight ?y))"


def test_construct_assignment_equations_with_simple_2d_equations_when_no_change_in_variables_returns_empty_set(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    for i in range(4):
        random_prev_load_limit = random.randint(0, 100)
        random_prev_current_load = random.randint(0, 100)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(random_prev_load_limit)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(random_prev_current_load)
        simple_prev_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_prev_state_fluents)
        load_action_state_fluent_storage.add_to_next_state_storage(simple_prev_state_fluents)

    (effects, numeric_preconditions, learned_perfectly,) = load_action_state_fluent_storage.construct_assignment_equations()
    assert learned_perfectly
    assert isinstance(effects, set)
    assert len(effects) == 0
    assert numeric_preconditions is None


def test_construct_assignment_equations_when_change_is_caused_by_constant_returns_correct_value(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    # This test is meant to validate that cases such as (assign (battery-level ?r) 10) can be handled.
    for i in range(3):
        random_prev_load_limit = random.randint(0, 100)
        random_prev_current_load = random.randint(0, 100)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(random_prev_load_limit)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(random_prev_current_load)
        simple_prev_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_prev_state_fluents)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(10)
        simple_next_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_next_state_storage(simple_next_state_fluents)

    (effects, numeric_preconditions, learned_perfectly,) = load_action_state_fluent_storage.construct_assignment_equations()
    assert learned_perfectly
    assert isinstance(effects, set)
    assert len(effects) == 1
    assert numeric_preconditions is None
    assert effects.pop().to_pddl() == "(assign (current_load ?z) 10)"


def test_construct_assignment_equations_with_simple_2d_equations_returns_correct_string_representation(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    previous_state_values = [(1, 7), (2, -1), (2, 14), (1, 0)]
    next_state_values = [9, 18, 18, 9]
    for prev_values, next_state_value in zip(previous_state_values, next_state_values):
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(prev_values[0])
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(prev_values[1])
        simple_prev_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_prev_state_fluents)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(next_state_value)
        simple_next_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_next_state_storage(simple_next_state_fluents)

    (effects, numeric_preconditions, learned_perfectly,) = load_action_state_fluent_storage.construct_assignment_equations()
    assert learned_perfectly
    assert isinstance(effects, set)
    assert len(effects) == 1
    assert numeric_preconditions is None
    assert effects.pop().to_pddl() == "(assign (current_load ?z) (* (load_limit ?z) 9))"


def test_construct_assignment_equations_with_two_equations_result_in_multiple_changes(load_action_state_fluent_storage: NumericFluentStateStorage,):
    previous_state_values = [(1, 7), (2, -1), (2, 14), (1, 0)]
    next_state_values = [(7, 9), (-16, 18), (14, 18), (-7, 9)]
    for prev_values, next_state_values in zip(previous_state_values, next_state_values):
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(prev_values[0])
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(prev_values[1])
        simple_prev_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_prev_state_fluents)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(next_state_values[0])
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(next_state_values[1])
        simple_next_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_next_state_storage(simple_next_state_fluents)

    (effects, numeric_preconditions, learned_perfectly,) = load_action_state_fluent_storage.construct_assignment_equations()
    assert learned_perfectly
    assert isinstance(effects, set)
    assert numeric_preconditions is None
    assignment_equations = {expression.to_pddl() for expression in effects}
    assert len(assignment_equations) == 2
    assert assignment_equations == {
        "(increase (load_limit ?z) (+ (* (load_limit ?z) -8) (* (current_load ?z) 2)))",
        "(assign (current_load ?z) (* (load_limit ?z) 9))",
    }


def test_construct_assignment_equations_with_an_increase_change_results_in_correct_values(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    previous_state_values = [(0, 7), (2, -1), (12, 32)]
    next_state_values = [(0, 8), (2, 0), (12, 33)]
    for prev_values, next_state_values in zip(previous_state_values, next_state_values):
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(prev_values[0])
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(prev_values[1])
        simple_prev_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_prev_state_fluents)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(next_state_values[0])
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(next_state_values[1])
        simple_next_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_next_state_storage(simple_next_state_fluents)

    (effects, numeric_preconditions, learned_perfectly,) = load_action_state_fluent_storage.construct_assignment_equations()
    assert learned_perfectly
    assert isinstance(effects, set)
    assert numeric_preconditions is None
    assignment_equations = {expression.to_pddl() for expression in effects}
    print(assignment_equations)


def test_construct_assignment_equations_with_fewer_equations_than_needed_to_create_single_solution_should_numeric_effect_and_preconditions(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    # Note the function that we are trying to calculate is y[i+1] = y[i] + 10 * x[i]
    # In this setting we create an observation where x[i] = 2 constantly but y[i] = 1, 2, 3, 4
    LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(2)
    for i in range(1, 10):
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(i)
        simple_prev_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_prev_state_fluents)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(i + 10 * 2)
        simple_next_state_fluents = {
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_next_state_storage(simple_next_state_fluents)

    (effects, numeric_preconditions, learned_perfectly,) = load_action_state_fluent_storage.construct_assignment_equations()
    assert learned_perfectly
    assert numeric_preconditions is not None
    assert effects is not None


def test_construct_safe_linear_inequalities_when_given_only_one_state_returns_degraded_conditions(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(411.0)
    CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(121.0)
    FUEL_COST_FUNCTION.set_value(34.0)
    simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
    }
    load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)
    output_conditions = load_action_state_fluent_storage.construct_safe_linear_inequalities(["(fuel-cost )", "(load_limit ?z)", "(current_load ?z)"])
    assert output_conditions.binary_operator == "and"
    assert len(output_conditions.operands) == 3
    numeric_operands = [operand for operand in output_conditions.operands if isinstance(operand, NumericalExpressionTree)]
    for op in numeric_operands:
        assert op.to_pddl() in ["(= (fuel-cost ) 34)", "(= (load_limit ?z) 411)", "(= (current_load ?z) 121)"]


def test_construct_safe_linear_inequalities_when_given_only_two_states_returns_smaller_convex_hull_and_const_condition(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(411.0)
    CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(121.0)
    FUEL_COST_FUNCTION.set_value(34.0)
    simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
    }
    load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)
    LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(413.0)
    CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(121.0)
    FUEL_COST_FUNCTION.set_value(35.0)
    another_simple_state_fluents = {
        "(fuel-cost )": FUEL_COST_FUNCTION,
        "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
    }
    load_action_state_fluent_storage.add_to_previous_state_storage(another_simple_state_fluents)
    output_conditions = load_action_state_fluent_storage.construct_safe_linear_inequalities(["(fuel-cost )", "(load_limit ?z)", "(current_load ?z)"])
    assert output_conditions.binary_operator == "and"
    assert len(output_conditions.operands) == 4
    numeric_conditions = [operand.to_pddl() for operand in output_conditions.operands if isinstance(operand, NumericalExpressionTree)]
    assert "(= (current_load ?z) 121)" in numeric_conditions  # the condition is shifted
    assert len([condition for condition in numeric_conditions if condition.startswith("(<=")]) == 2
    assert len([condition for condition in numeric_conditions if condition.startswith("(=")]) == 2
    print(str(output_conditions))


def test_construct_safe_linear_inequalities_will_create_correct_inequalities_when_given_three_points_for_two_variables(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    pre_state_input_values = [(0.0, 0.0), (0.0, 1.0), (1.0, 0.0)]
    for fuel_cost_val, current_limit_val in pre_state_input_values:
        FUEL_COST_FUNCTION.set_value(fuel_cost_val)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(current_limit_val)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(0.0)
        simple_state_fluents = {
            "(fuel-cost )": FUEL_COST_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)

    output_conditions = load_action_state_fluent_storage.construct_safe_linear_inequalities()

    expected_conditions = [
        "(<= (* (fuel-cost ) -1) 0)",
        "(<= (* (current_load ?z) -1) 0)",
        "(<= (+ (* (current_load ?z) 0.7071) (* (fuel-cost ) 0.7071)) 0.7071)",
        "(= (load_limit ?z) 0)",
    ]
    for _, precondition in output_conditions:
        assert precondition.to_pddl() in expected_conditions


def test_construct_safe_linear_inequalities_will_create_correct_inequalities_when_given_four_points_for_two_variables(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    pre_state_input_values = [(0.0, 0.0), (1.0, 0.0), (1.0, 1.0), (0.0, 1.0)]
    for fuel_cost_val, current_limit_val in pre_state_input_values:
        FUEL_COST_FUNCTION.set_value(fuel_cost_val)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(current_limit_val)
        LOAD_LIMIT_TRAJECTORY_FUNCTION.set_value(0.0)
        simple_state_fluents = {
            "(fuel-cost )": FUEL_COST_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
            "(load_limit ?z)": LOAD_LIMIT_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)

    output_conditions = load_action_state_fluent_storage.construct_safe_linear_inequalities()
    expected_conditions = [
        "(<= (current_load ?z) 1)",
        "(<= (* (fuel-cost ) -1) 0)",
        "(<= (fuel-cost ) 1)",
        "(<= (* (current_load ?z) -1) 0)",
        "(= (load_limit ?z) 0)",
    ]

    for _, precondition in output_conditions:
        assert precondition.to_pddl() in expected_conditions


def test_construct_safe_linear_inequalities_with_one_dimension_variable_select_min_and_max_values(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):

    pre_state_input_values = [-19.0, 14.0, 28.0, -7.0]
    for fuel_cost_val in pre_state_input_values:
        FUEL_COST_FUNCTION.set_value(fuel_cost_val)
        simple_state_fluents = {
            "(fuel-cost )": FUEL_COST_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)

    output_conditions = load_action_state_fluent_storage.construct_safe_linear_inequalities()
    expected_conditions = [
        "(<= (fuel-cost ) 28)",
        "(>= (fuel-cost ) -19)",
    ]
    for _, precondition in output_conditions:
        assert precondition.to_pddl() in expected_conditions


def test_construct_safe_linear_inequalities_when_not_given_relevant_fluents_uses_all_variables_in_previous_state(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    pre_state_input_values = [(-19.0, 32.0), (14.0, 52.0), (28.0, 12.0), (-7.0, 13.0)]
    for fuel_cost_val, current_limit_val in pre_state_input_values:
        FUEL_COST_FUNCTION.set_value(fuel_cost_val)
        CURRENT_LOAD_TRAJECTORY_FUNCTION.set_value(current_limit_val)

        simple_state_fluents = {
            "(fuel-cost )": FUEL_COST_FUNCTION,
            "(current_load ?z)": CURRENT_LOAD_TRAJECTORY_FUNCTION,
        }
        load_action_state_fluent_storage.add_to_previous_state_storage(simple_state_fluents)

    output_conditions = load_action_state_fluent_storage.construct_safe_linear_inequalities()
    print(output_conditions)
    for _, precondition in output_conditions:
        assert "(fuel-cost )" in precondition.to_pddl()
        assert "(current_load ?z)" in precondition.to_pddl()


def test_detect_linear_dependent_features_when_given_only_one_sample_returns_that_no_column_is_dependent_since_there_are_not_enough_samples_to_guarantee_correctness():
    pre_state_data = {
        "(fuel-cost )": [1.0],
        "(current_load ?z)": [1.0],
    }
    linear_dependant_matrix = DataFrame(pre_state_data)
    output_matrix, linear_dependent_fluents, remained_fluents = detect_linear_dependent_features(linear_dependant_matrix)
    assert list(output_matrix.columns) == ["(fuel-cost )", "(current_load ?z)"]


def test_detect_linear_dependent_features_detects_that_two_equal_features_are_linear_dependant():
    pre_state_data = {
        "(fuel-cost )": [1.0, 2.0],
        "(current_load ?z)": [1.0, 2.0],
    }
    linear_dependant_matrix = DataFrame(pre_state_data)

    output_matrix, linear_dependent_fluent_strs, removed_fluents = detect_linear_dependent_features(linear_dependant_matrix)

    assert linear_dependent_fluent_strs == ["(= (current_load ?z) (* 1 (fuel-cost )))"]
    assert list(output_matrix.columns) == ["(fuel-cost )"]
    assert removed_fluents == {"(current_load ?z)": "(fuel-cost )"}


def test_detect_linear_dependent_features_detects_that_two_linear_dependent_features_are_linear_dependant(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    pre_state_data = {
        "(fuel-cost )": [1.0, 2.0],
        "(current_load ?z)": [2.0, 4.0],
    }
    linear_dependant_matrix = DataFrame(pre_state_data)
    output_matrix, linear_dependent_fluent_strs, removed_fluents = detect_linear_dependent_features(linear_dependant_matrix)

    assert linear_dependent_fluent_strs == ["(= (current_load ?z) (* 2 (fuel-cost )))"]
    assert removed_fluents == {"(current_load ?z)": "(fuel-cost )"}
    assert output_matrix.shape[1] == 1


def test_detect_linear_dependent_features_detects_when_variables_are_independent():
    pre_state_data = {
        "(fuel-cost )": [-19.0, 14.0, 28],
        "(current_load ?z)": [32.0, 54.0, 17.0],
    }
    linear_dependant_matrix = DataFrame(pre_state_data)
    relevant_fluents = ["(fuel-cost )", "(current_load ?z)"]
    output_matrix, linear_dependent_fluent_strs, removed_fluents_map = detect_linear_dependent_features(linear_dependant_matrix)

    assert linear_dependent_fluent_strs == []
    assert removed_fluents_map == {}
    assert output_matrix.shape[1] == 2


def test_detect_linear_dependent_features_when_one_feature_is_zero_constant_does_not_include_it_in_the_additional_conditions():
    pre_state_data = {
        "(fuel-cost )": [0.0, 0.0, 0.0],
        "(current_load ?z)": [32.0, 54.0, 17.0],
    }
    linear_dependant_matrix = DataFrame(pre_state_data)
    output_matrix, linear_dependent_fluent_strs, removed_fluents_map = detect_linear_dependent_features(linear_dependant_matrix)

    assert linear_dependent_fluent_strs == []
    assert removed_fluents_map == {}
    assert output_matrix.shape[1] == 2


def test_filter_constant_features_detects_features_that_equal_to_a_single_constant_value(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    pre_state_data = {
        "(fuel-cost )": [1.0, 32.0, 95.0],
        "(current_load ?z)": [15.0, 12.0, 17.0],
        "(load_limit ?z)": [33.0, 33.0, 33.0],
    }
    matrix_with_constant_column = DataFrame(pre_state_data)
    filtered_matrix, equal_fluent_strs, removed_fluents = filter_constant_features(matrix_with_constant_column)
    assert filtered_matrix.shape[1] == 2
    assert equal_fluent_strs == ["(= (load_limit ?z) 33.0)"]
    assert removed_fluents == ["(load_limit ?z)"]


def test_filter_constant_features_with_two_constant_detects_both_and_removes_them(load_action_state_fluent_storage: NumericFluentStateStorage,):
    pre_state_data = {
        "(fuel-cost )": [1.0, 32.0, 95.0],
        "(current_load ?z)": [15.0, 12.0, 17.0],
        "(load_limit ?z)": [33.0, 33.0, 33.0],
        "(weight ?z)": [5.0, 5.0, 5.0],
    }
    matrix_with_constant_column = DataFrame(pre_state_data)
    filtered_matrix, equal_fluent_strs, removed_fluents = filter_constant_features(matrix_with_constant_column)
    assert filtered_matrix.shape[1] == 2
    assert set(equal_fluent_strs) == {"(= (load_limit ?z) 33.0)", "(= (weight ?z) 5.0)"}
    assert set(removed_fluents) == {"(load_limit ?z)", "(weight ?z)"}


def test_filter_constant_features_with_all_features_being_constant_returns_correct_empty_df_and_correct_conditions(
    load_action_state_fluent_storage: NumericFluentStateStorage,
):
    pre_state_data = {
        "(fuel-cost )": [1.0, 1.0, 1.0],
        "(current_load ?z)": [15.0, 15.0, 15.0],
        "(load_limit ?z)": [33.0, 33.0, 33.0],
        "(weight ?z)": [5.0, 5.0, 5.0],
    }
    matrix_with_constant_column = DataFrame(pre_state_data)
    filtered_matrix, equal_fluent_strs, removed_fluents = filter_constant_features(matrix_with_constant_column)
    assert filtered_matrix.shape[1] == 0
    assert filtered_matrix.shape[0] == 0

    assert set(equal_fluent_strs) == {
        "(= (load_limit ?z) 33.0)",
        "(= (weight ?z) 5.0)",
        "(= (fuel-cost ) 1.0)",
        "(= (current_load ?z) 15.0)",
    }
    assert set(removed_fluents) == {"(load_limit ?z)", "(weight ?z)", "(fuel-cost )", "(current_load ?z)"}



"""Constants for the tests."""

from pathlib import Path
from typing import Dict, List

from pddl_plus_parser.exporters.numeric_trajectory_exporter import parse_action_call
from pddl_plus_parser.models import (
    PDDLType,
    Predicate,
    PDDLFunction,
    ObservedComponent,
    PDDLObject,
    MultiAgentComponent,
    ActionCall,
    CompoundPrecondition,
)

from sam_learning.learners import SAMLearner

EXAMPLES_DIR_PATH = Path(__file__).parent / "examples"
LOGISTICS_DOMAIN_PATH = EXAMPLES_DIR_PATH / "domain-logistics.pddl"
WOODWORKING_DOMAIN_PATH = EXAMPLES_DIR_PATH / "woodworking-domain.pddl"
WOODWORKING_PROBLEM_PATH = EXAMPLES_DIR_PATH / "woodworking_problem.pddl"
WOODWORKING_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "woodworking_trajectory.trajectory"

DEPOTS_NUMERIC_DOMAIN_PATH = EXAMPLES_DIR_PATH / "depot_numeric.pddl"
DEPOTS_NUMERIC_PROBLEM_PATH = EXAMPLES_DIR_PATH / "pfile2.pddl"

DEPOTS_DISCRETE_DOMAIN_PATH = EXAMPLES_DIR_PATH / "depot_discrete.pddl"
DEPOTS_DISCRETE_PROBLEM_PATH = EXAMPLES_DIR_PATH / "depot_discrete_problem.pddl"

ELEVATORS_DOMAIN_PATH = EXAMPLES_DIR_PATH / "elevators_domain.pddl"
ELEVATORS_PROBLEM_PATH = EXAMPLES_DIR_PATH / "elevators_p03.pddl"
ELEVATORS_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "test_trajectory"
DEPOT_NUMERIC_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "test_numeric_trajectory"
DEPOT_FLUENTS_MAP_PATH = EXAMPLES_DIR_PATH / "depot_fluents_map.json"

SATELLITE_DOMAIN_PATH = EXAMPLES_DIR_PATH / "metricSat.pddl"
SATELLITE_PROBLEM_PATH = EXAMPLES_DIR_PATH / "pfile3.pddl"
SATELLITE_NUMERIC_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "pfile3.trajectory"

SATELLITE_PROBLEMATIC_PROBLEM_PATH = EXAMPLES_DIR_PATH / "pfile1_sat.pddl"
SATELLITE_PROBLEMATIC_NUMERIC_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "pfile1_sat.trajectory"

SATELLITE_FLUENTS_MAP_PATH = EXAMPLES_DIR_PATH / "satellite_fluents_map.json"

SAILING_EXPECTED_DOMAIN_PATH = EXAMPLES_DIR_PATH / "sailing_expected_domain.pddl"
SAILING_LEARNED_DOMAIN_PATH = EXAMPLES_DIR_PATH / "sailing_domain.pddl"
SAILING_PROBLEM_PATH = EXAMPLES_DIR_PATH / "prob_sailing.pddl"
SAILING_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "sailing_observation.trajectory"

WOODWORKING_COMBINED_DOMAIN_PATH = EXAMPLES_DIR_PATH / "woodworking_combined_domain.pddl"
WOODWORKING_COMBINED_PROBLEM_PATH = EXAMPLES_DIR_PATH / "woodworking_combined_problem.pddl"
WOODWORKING_COMBINED_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "ma_woodworking_trajectory.trajectory"

ROVERS_COMBINED_DOMAIN_PATH = EXAMPLES_DIR_PATH / "rover_combined_domain.pddl"
ROVERS_ESAM_DOMAIN_PATH = EXAMPLES_DIR_PATH / "rovers_esam_domain.pddl"
ROVERS_COMBINED_PROBLEM_PATH = EXAMPLES_DIR_PATH / "rovers_conflicing_actions_problem.pddl"
ROVERS_COMBINED_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "rovers_conflicting_actions_trajectory.trajectory"
ROVERS_COMBINED_ESAM_PROBLEM_PATH = EXAMPLES_DIR_PATH / "rovers_esam_problem.pddl"
ROVERS_COMBINED_ESAM_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "rovers_esam_problem.trajectory"

SPIDER_DOMAIN_PATH = EXAMPLES_DIR_PATH / "spider_domain.pddl"
SPIDER_PROBLEM_PATH = EXAMPLES_DIR_PATH / "spider_problem.pddl"
SPIDER_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "spider_trajectory.trajectory"

NURIKABE_DOMAIN_PATH = EXAMPLES_DIR_PATH / "nurikabe_domain.pddl"
NURIKABE_PROBLEM_PATH = EXAMPLES_DIR_PATH / "nurikabe_pfile_random-9x9-107.pddl"
NURIKABE_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "nurikabe_pfile_random-9x9-107.trajectory"

ADL_SATELLITE_DOMAIN_PATH = EXAMPLES_DIR_PATH / "adlSat.pddl"
ADL_SATELLITE_PROBLEM_PATH = EXAMPLES_DIR_PATH / "adlSat_problem.pddl"
ADL_SATELLITE_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "adlSat_trajectory.trajectory"

MINECRAFT_DOMAIN_PATH = EXAMPLES_DIR_PATH / "minecraft_domain.pddl"
MINECRAFT_PROBLEM_PATH = EXAMPLES_DIR_PATH / "minecraft_pfile0.pddl"
MINECRAFT_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "minecraft_pfile0.trajectory"
MINECRAFT_FLUENTS_MAP_PATH = EXAMPLES_DIR_PATH / "minecraft_fluents_map.json"

STAR_CRAFT_DOMAIN_PATH = EXAMPLES_DIR_PATH / "starcraft_domain.pddl"
STAR_CRAFT_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "starcraft_trajectory.trajectory"
STAR_CRAFT_FLUENTS_MAP_PATH = EXAMPLES_DIR_PATH / "starcraft_fluents_map.json"

MINECRAFT_MEDIUM_DOMAIN_PATH = EXAMPLES_DIR_PATH / "domain_minecraft_medium.pddl"
MINECRAFT_MEDIUM_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "trajectory_minecraft_medium.trajectory"
MINECRAFT_MEDIUM_FLUENTS_MAP_PATH = EXAMPLES_DIR_PATH / "fluents_map_minecraft_medium.json"

MINECRAFT_SMALL_DOMAIN_PATH = EXAMPLES_DIR_PATH / "minecraft_domain_small.pddl"
MINECRAFT_SMALL_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "minecraft_problem_small.trajectory"

MINECRAFT_LARGE_DOMAIN_PATH = EXAMPLES_DIR_PATH / "advanced_minecraft_domain.pddl"
MINECRAFT_LARGE_PROBLEM_PATH = EXAMPLES_DIR_PATH / "advanced_map_instance_0.pddl"
MINECRAFT_LARGE_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "advanced_map_instance_0.trajectory"

DRIVERLOG_POLY_DOMAIN_PATH = EXAMPLES_DIR_PATH / "driverlogHardNumeric.pddl"
DRIVERLOG_POLY_PROBLEM_PATH = EXAMPLES_DIR_PATH / "driverlog_poly_problem.pddl"
DRIVERLOG_POLY_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "driverlog_poly_problem.trajectory"

FARMLAND_DOMAIN_PATH = EXAMPLES_DIR_PATH / "farmland.pddl"
FARMLAND_TRAJECTORIES_DIRECTORY = EXAMPLES_DIR_PATH / "large_data_examples" / "farmland"
SAILING_TRAJECTORIES_DIRECTORY = EXAMPLES_DIR_PATH / "large_data_examples" / "sailing"

COUNTERS_POLYNOMIAL_DOMAIN_PATH = EXAMPLES_DIR_PATH / "counters_poly.pddl"
COUNTERS_POLYNOMIAL_PROBLEMS_PATH = [path for path in EXAMPLES_DIR_PATH.glob("pfile_counters_poly*.pddl")]
COUNTERS_POLYNOMIAL_TRAJECTORIES_PATH = [path for path in EXAMPLES_DIR_PATH.glob("pfile_counters_poly*.trajectory")]

BLOCKS_PROPOSITIONAL_DOMAIN_PATH = EXAMPLES_DIR_PATH / "blocksworld_propositional_domain.pddl"
BLOCKS_PROPOSITIONAL_PROBLEM_PATH = EXAMPLES_DIR_PATH / "blocks_propositional_prob00.pddl"
BLOCKS_PROPOSITIONAL_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "blocks_propositional.trajectory"

TEST_PPO_OBSERVATIONS_DIRECTORY = EXAMPLES_DIR_PATH / "ppo_observations"
TEST_PPO_MINECRAFT_DOMAIN = EXAMPLES_DIR_PATH / "ppo_minecraft_domain.pddl"

BARMAN_CLASSICAL_DOMAIN_PATH = EXAMPLES_DIR_PATH / "barman_domain.pddl"
BARMAN_CLASSICAL_PROBLEM_PATH = EXAMPLES_DIR_PATH / "barman_pfile3.pddl"
BARMAN_CLASSICAL_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "barman_pfile3.trajectory"

FARMLAND_EXAMPLES_PATH = EXAMPLES_DIR_PATH / "convex_state_data.csv"
FARMLAND_PAPER_EXAMPLES_PATH = EXAMPLES_DIR_PATH / "convex_state_data_for_paper.csv"
FARMLAND_SPAN_EXAMPLES_PATH = EXAMPLES_DIR_PATH / "span_convex_state_data.csv"

DEPOT_ONLINE_LEARNING_PROBLEM = EXAMPLES_DIR_PATH / "depot_numeric_problem_online_learning.pddl"
DEPOT_ONLINE_LEARNING_PROBLEM_WITH_NUMERIC_GOAL = (
        EXAMPLES_DIR_PATH / "depot_numeric_problem_online_learning_with_numeric_goals.pddl"
)
DEPOT_ONLINE_LEARNING_PLAN = EXAMPLES_DIR_PATH / "depot_numeric_problem_online_learning.solution"

OBJECT_TYPE = PDDLType(name="object")
AGENT_TYPE = PDDLType(name="agent")
CITY_TYPE = PDDLType(name="city", parent=OBJECT_TYPE)
WOODOBJ_TYPE = PDDLType(name="woodobj", parent=OBJECT_TYPE)
SURFACE_TYPE = PDDLType(name="surface", parent=OBJECT_TYPE)
TREATMENT_STATUS_TYPE = PDDLType(name="treatmentstatus", parent=OBJECT_TYPE)
COLOR_TYPE = PDDLType(name="acolour", parent=OBJECT_TYPE)
PART_TYPE = PDDLType(name="part", parent=WOODOBJ_TYPE)
TAXI_TYPE = PDDLType(name="taxi", parent=AGENT_TYPE)
TRUCK_TYPE = PDDLType(name="truck", parent=AGENT_TYPE)
CRATE_TYPE = PDDLType(name="crate", parent=OBJECT_TYPE)
AIRPLANE_TYPE = PDDLType(name="airplane", parent=AGENT_TYPE)
LOCATION_TYPE = PDDLType(name="location", parent=OBJECT_TYPE)
COUNT_TYPE = PDDLType(name="count", parent=OBJECT_TYPE)

AT_TRUCK_PREDICATE = Predicate(name="at", signature={"?a": AGENT_TYPE, "?loc": LOCATION_TYPE})

FUEL_COST_FUNCTION = PDDLFunction(name="fuel-cost", signature={})
LOAD_LIMIT_TRAJECTORY_FUNCTION = PDDLFunction(name="load_limit", signature={"?z": TRUCK_TYPE})
LOAD_LIMIT_GROUNDED_TRAJECTORY_FUNCTION = PDDLFunction(name="load_limit", signature={"truck1": TRUCK_TYPE})
CURRENT_LOAD_TRAJECTORY_FUNCTION = PDDLFunction(name="current_load", signature={"?z": TRUCK_TYPE})
CURRENT_LOAD_GROUNDED_TRAJECTORY_FUNCTION = PDDLFunction(name="current_load", signature={"truck1": TRUCK_TYPE})
WEIGHT_FUNCTION = PDDLFunction(name="weight", signature={"?c": CRATE_TYPE})

DRIVERLOG_COMBINED_DOMAIN_PATH = EXAMPLES_DIR_PATH / "driverlog_combined_domain.pddl"
DRIVERLOG_COMBINED_PROBLEM_PATH = EXAMPLES_DIR_PATH / "driverlog_conflicting_problem.pddl"
DRIVERLOG_COMBINED_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "driverlog_ma_sam_pfile2.trajectory"
ROVERS_COMBINED_WITH_MACRO_DOMAIN_PATH = EXAMPLES_DIR_PATH / "rovers_ma_plus_domain.pddl"

TWO_SIDES_OF_BOX_PATH = EXAMPLES_DIR_PATH / "negative_from_two_sides.csv"
CLOSE_TO_BOX_PATH = EXAMPLES_DIR_PATH / "negative_close_to_box.csv"
CLOSE_TO_LINEAR_CONDITION_PATH = EXAMPLES_DIR_PATH / "negative_close_to_linear_condition.csv"

COUNTERS_ONLINE_LEARNING_DOMAIN_PATH = EXAMPLES_DIR_PATH / "counters_domain_online_learning.pddl"
COUNTERS_ONLINE_LEARNING_PROBLEM_PATH = EXAMPLES_DIR_PATH / "pfile10_0_counters_online.pddl"

DEPOT_ERROR_IN_INCREMENTAL_HULL_CSV = EXAMPLES_DIR_PATH / "depot_numeric_unload_action_error_case.csv"

BARMAN_ERROR_IN_PRODUCTION_DOMAIN_PATH = EXAMPLES_DIR_PATH / "barman_bug_in_prod.pddl"
BARMAN_ERROR_IN_PRODUCTION_PROBLEM_PATH = EXAMPLES_DIR_PATH / "0_barman_prob.pddl"
BARMAN_ERROR_IN_PRODUCTION_TRAJECTORY_PATH = EXAMPLES_DIR_PATH / "0_barman_traj.trajectory"


def sync_snapshot(
        sam_learning: SAMLearner,
        component: ObservedComponent,
        trajectory_objects: Dict[str, PDDLObject],
        should_include_all_objects: bool = False,
) -> None:
    previous_state = component.previous_state
    next_state = component.next_state
    test_action_call = component.grounded_action_call
    sam_learning.current_trajectory_objects = trajectory_objects
    if should_include_all_objects:
        all_types = [pddl_type for pddl_type in sam_learning.partial_domain.types if pddl_type != "object"]
    else:
        all_types = []

    sam_learning.triplet_snapshot.create_triplet_snapshot(
        previous_state=previous_state,
        next_state=next_state,
        current_action=test_action_call,
        observation_objects=trajectory_objects,
        specific_types=all_types,
    )


def extract_preconditions_predicates(compound_preconditions: CompoundPrecondition) -> List[Predicate]:
    predicates = []
    for _, precond in compound_preconditions:
        if isinstance(precond, Predicate):
            predicates.append(precond)

    return predicates


def create_plan_actions(plan_path: Path) -> List[ActionCall]:
    """Creates a list of action calls from a plan file.

    :param plan_path:
    :return:
    """
    with open(plan_path, "rt") as plan_file:
        plan_lines = plan_file.readlines()

    plan_actions = []
    for line in plan_lines:
        plan_actions.append(parse_action_call(line))

    return plan_actions



"""Module responsible for calculating our approach for Multi-Agent precision and recall."""
import csv
from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Tuple

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import (
    Domain,
    MultiAgentObservation,
)
from pddl_plus_parser.multi_agent import MultiAgentTrajectoryExporter

from statistics import SemanticPerformanceCalculator, _calculate_precision_recall
from utilities import LearningAlgorithmType, NegativePreconditionPolicy

SEMANTIC_PRECISION_STATS = [
    "fold",
    "learning_algorithm",
    "policy",
    "action_name",
    "num_trajectories",
    "precondition_precision",
    "precondition_recall",
]


class MASamPerformanceCalculator(SemanticPerformanceCalculator):
    """Class responsible for calculating the Multi-Agent precision and recall of a model."""

    dataset_observations: List[MultiAgentObservation]

    def __init__(
        self,
        model_domain: Domain,
        model_domain_path: Path,
        observations: List[MultiAgentObservation],
        working_directory_path: Path,
        learning_algorithm: LearningAlgorithmType,
    ):
        super().__init__(model_domain, model_domain_path, observations, working_directory_path, learning_algorithm)

    def calculate_preconditions_semantic_performance(
        self, learned_domain: Domain, learned_domain_path: Path
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Calculates the precision recall values of the learned preconditions.

        :param learned_domain: the action model that was learned using the action model learning algorithm
        :param learned_domain_path: the path to the learned domain.
        :return: the precision and recall dictionaries.
        """
        num_true_positives = defaultdict(int)
        num_false_negatives = defaultdict(int)
        num_false_positives = defaultdict(int)
        self.logger.debug("Starting to calculate the semantic preconditions performance")
        for index, observation in enumerate(self.dataset_observations):
            observation_objects = observation.grounded_objects
            for component in observation.components:
                for action in component.grounded_joint_action.actions:
                    if action.name not in learned_domain.actions:
                        continue

                    true_positive, false_positive, false_negative = self._calculate_action_applicability_rate(
                        action, learned_domain_path, component.previous_state, observation_objects,
                    )
                    num_true_positives[action.name] += true_positive
                    num_false_positives[action.name] += false_positive
                    num_false_negatives[action.name] += false_negative

        return _calculate_precision_recall(num_false_negatives, num_false_positives, num_true_positives)

    def _calculate_effects_difference_rate_for_multi_agent_experiments(
        self,
        observation: MultiAgentObservation,
        learned_domain: Domain,
        ma_trajectory_exporter: MultiAgentTrajectoryExporter,
        num_false_negatives: Dict[str, int],
        num_false_positives: Dict[str, int],
        num_true_positives: Dict[str, int],
    ) -> None:
        """Calculates the effects difference rate for each action.

        :param observation: the observation that is being tested.
        :param learned_domain: the domain that was learned using the action model learning algorithm.
        :param ma_trajectory_exporter: module able to create multi-agent trajectories.
        :param num_false_negatives: the dictionary mapping between the action name and the number of false negative
        :param num_false_positives: the dictionary mapping between the action name and the number of false positive
        :param num_true_positives: the dictionary mapping between the action name and the number of true positive
        """
        self.logger.info("Calculating effects difference rate for the observation.")
        for observation_triplet in observation.components:
            model_previous_state = observation_triplet.previous_state
            executed_joint_action = observation_triplet.grounded_joint_action
            model_next_state = observation_triplet.next_state
            if any([action.name not in learned_domain.actions for action in executed_joint_action.actions]):
                continue

            try:
                ma_triplet = ma_trajectory_exporter.create_multi_agent_triplet(
                    previous_state=model_previous_state,
                    action_call=str(executed_joint_action),
                    problem_objects=observation.grounded_objects,
                    allow_inapplicable_actions=False,
                )

                model_state_predicates = {
                    predicate.untyped_representation
                    for grounded_predicate in model_next_state.state_predicates.values()
                    for predicate in grounded_predicate
                }
                learned_state_predicates = {
                    predicate.untyped_representation
                    for grounded_predicate in ma_triplet.next_state.state_predicates.values()
                    for predicate in grounded_predicate
                }

                for action in observation_triplet.grounded_joint_action.actions:
                    num_true_positives[action.name] += len(model_state_predicates.intersection(learned_state_predicates))
                    num_false_positives[action.name] += len(learned_state_predicates.difference(model_state_predicates))
                    num_false_negatives[action.name] += len(model_state_predicates.difference(learned_state_predicates))

            except ValueError:
                self.logger.debug("The joint action is not applicable according to the learned model.")
                continue

    def calculate_effects_semantic_performance(self, learned_domain: Domain) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Calculates the precision recall values of the learned effects.

        :param learned_domain: the action model that was learned using the action model learning algorithm
        :return: the precision and recall dictionaries.
        """
        self.logger.info("Starting to calculate the semantic effects performance for the multi-agent model.")
        num_true_positives = defaultdict(int)
        num_false_negatives = defaultdict(int)
        num_false_positives = defaultdict(int)
        if len(self.model_domain.predicates) == 0:
            return (
                {action_name: 1 for action_name in self.model_domain.actions.keys()},
                {action_name: 1 for action_name in self.model_domain.actions.keys()},
            )

        multi_agent_trajectory_exporter = MultiAgentTrajectoryExporter(domain=learned_domain, allow_invalid_actions=False)
        for observation in self.dataset_observations:
            self._calculate_effects_difference_rate_for_multi_agent_experiments(
                observation, learned_domain, multi_agent_trajectory_exporter, num_false_negatives, num_false_positives, num_true_positives
            )

        return _calculate_precision_recall(num_false_negatives, num_false_positives, num_true_positives)

    def calculate_performance_for_ma_sam_experiments(
        self, learned_domain_path: Path, num_used_observations: int, policy: NegativePreconditionPolicy
    ) -> None:
        """Calculates the model's performance with both the precision and the recall values calculated.
        :param learned_domain_path: the path to the learned action model.
        :param num_used_observations: the number of observations used to learn the action model.
        :param policy: the negative precondition policy to use.
        """
        learned_domain = DomainParser(domain_path=learned_domain_path, partial_parsing=False).parse_domain()
        self.logger.info("Starting to calculate the semantic preconditions performance of the learned domain.")
        preconditions_precision, preconditions_recall = self.calculate_preconditions_semantic_performance(learned_domain, learned_domain_path)
        self.logger.info("Starting to calculate the semantic effects performance of the learned domain.")
        effects_precision, effects_recall = self.calculate_effects_semantic_performance(learned_domain)
        for action_name in self.model_domain.actions:
            action_stats = {
                "learning_algorithm": self.learning_algorithm.name,
                "policy": policy.name,
                "action_name": action_name,
                "num_trajectories": num_used_observations,
                "precondition_precision": preconditions_precision.get(action_name, 1),
                "precondition_recall": preconditions_recall.get(action_name, 0),
                "effects_precision": effects_precision.get(action_name, 1),
                "effects_recall": effects_recall.get(action_name, 1),
            }
            self.combined_stats.append(action_stats)

        self.logger.info(f"Finished calculating the learning performance for {self.learning_algorithm.name}.")

    def export_combined_semantic_performance(self) -> None:
        """Export the numeric learning statistics to a CSV report file."""
        statistics_path = self.results_dir_path / f"{self.learning_algorithm.name}_{self.model_domain.name}_combined_semantic_performance.csv"
        with open(statistics_path, "wt", newline="") as statistics_file:
            stats_writer = csv.DictWriter(statistics_file, fieldnames=SEMANTIC_PRECISION_STATS)
            stats_writer.writeheader()
            for data_line in self.combined_stats:
                stats_writer.writerow(data_line)



"""Create a train and test set split for the input directory for the action model learning algorithms to use."""
import json
import logging
import random
import shutil
from pathlib import Path
from typing import Tuple, List, Iterator, Generator, Dict

from sklearn.model_selection import train_test_split, KFold

RANDOM_STATE = 42  # it is always good to seed something according to the most important number in the world!
DEFAULT_TEST_SIZE = 0.2

FOLDS_CONFIG_FILE_NAME = "folds.json"
FOLDS_LABEL = "fold"


def create_test_set_indices(array_size: int, n_split: int, only_train_test: bool = False) -> Iterator[Tuple[List[int], List[int]]]:
    """Creates the indices to use to create the train and the test set of the problems.

    :param array_size: the size of the array containing the problems.
    :param n_split: the number of splits to divide the dataset to.
    :param only_train_test: whether to only split to train and test.
    :return: the train and the test set indices for each fold.
    """
    indices = list(range(array_size))
    if not only_train_test:
        kf = KFold(n_splits=n_split, random_state=RANDOM_STATE, shuffle=True)
        for train, test in kf.split(indices):
            yield train, test

    else:
        train, test = train_test_split(indices, test_size=DEFAULT_TEST_SIZE, random_state=RANDOM_STATE, shuffle=True)
        yield train, test


def save_fold_settings(workdir_path: Path, folds_configuration: Dict[str, Dict[str, List[str]]]) -> None:
    """Saves the fold settings to a json file.

    :param workdir_path: the path to the working directory.
    :param folds_configuration: the folds configuration, i.e. the train and test problems.
    """
    with open(workdir_path / FOLDS_CONFIG_FILE_NAME, "wt") as folds_file:
        json.dump(folds_configuration, folds_file)


def load_fold_settings(workdir_path: Path) -> Dict[str, Dict[str, List[Path]]]:
    """Loads the fold settings from a json file.

    :param workdir_path: the path to the working directory.
    :return: the folds configuration, i.e. the train and test problems.
    """
    if not (workdir_path / FOLDS_CONFIG_FILE_NAME).exists():
        return {}

    folds_configurations = {}
    with open(workdir_path / FOLDS_CONFIG_FILE_NAME, "rt") as folds_file:
        config_data = json.load(folds_file)
        for fold_name, fold in config_data.items():
            folds_configurations[fold_name] = {
                "train": [Path(problem) for problem in fold["train"]],
                "test": [Path(problem) for problem in fold["test"]],
            }

    return folds_configurations


def load_distributed_fold_settings(workdir_path: Path) -> Dict[str, Dict[str, List[Path]]]:
    """Loads the fold settings from a json file.

    :param workdir_path: the path to the working directory.
    :return: the folds configuration, i.e. the train and test problems.
    """
    if not (workdir_path / FOLDS_CONFIG_FILE_NAME).exists():
        return {}

    folds_configurations = {}
    with open(workdir_path / FOLDS_CONFIG_FILE_NAME, "rt") as folds_file:
        config_data = json.load(folds_file)
        for fold_name, fold in config_data.items():
            folds_configurations[fold_name] = {"train": fold["train"], "test": [Path(problem) for problem in fold["test"]]}

    return folds_configurations


class KFoldSplit:
    """Split the working directory into a train and test set directories."""

    logger: logging.Logger
    working_directory_path: Path
    n_split: int
    train_set_dir_path: Path
    test_set_dir_path: Path
    domain_file_path: Path
    only_train_test: bool

    def __init__(self, working_directory_path: Path, domain_file_name: str, n_split: int = 0, only_train_test: bool = False):
        self.logger = logging.getLogger(__name__)
        self.working_directory_path = working_directory_path
        self.n_split = n_split
        self.train_set_dir_path = working_directory_path / "train"
        self.test_set_dir_path = working_directory_path / "test"
        self.domain_file_path = working_directory_path / domain_file_name
        self.only_train_test = only_train_test

    def _copy_domain(self) -> None:
        """Copies the domain to the train set directory so that it'd be used in the learning process."""
        self.logger.debug("Copying the domain to the train set directory.")
        shutil.copy(self.domain_file_path, self.train_set_dir_path / self.domain_file_path.name)

    def remove_created_directories(self) -> None:
        """Deletes the train and test directories."""
        self.logger.debug("Deleting the train set directory!")
        shutil.rmtree(self.train_set_dir_path)
        self.logger.debug("Deleting the test set directory!")
        shutil.rmtree(self.test_set_dir_path)

    def create_directories_content(self, train_set_problems: List[Path], test_set_problems: List[Path], trajectory_paths: List[Path]) -> None:
        self.train_set_dir_path.mkdir(exist_ok=True)
        self.test_set_dir_path.mkdir(exist_ok=True)
        self._copy_domain()

        for problem in test_set_problems:
            shutil.copy(problem, self.test_set_dir_path / problem.name)

        for trajectory, problem in zip(trajectory_paths, train_set_problems):
            shutil.copy(trajectory, self.train_set_dir_path / trajectory.name)
            shutil.copy(problem, self.train_set_dir_path / problem.name)

    def create_k_fold(
        self, trajectory_suffix: str = "*.trajectory", max_items: int = 0, load_configuration: bool = True
    ) -> Generator[Tuple[Path, Path], None, None]:
        """Creates a generator that will be used for the next algorithm to know where the train and test set
            directories reside.

        :param trajectory_suffix: the suffix of the trajectory files to be used.
        :param max_items: the maximum number of items to be used in the train and test set together.
        :param load_configuration: whether to load folds settings from the configuration file.
        :return: a generator for the train and test set directories.
        """
        self.logger.info("Starting to create the folds for the cross validation process.")
        folds_data = load_fold_settings(self.working_directory_path)
        if load_configuration and len(folds_data) > 0:
            self.logger.info("Loading the folds settings from the configuration file.")
            for fold_name, fold_content in folds_data.items():
                train_set_trajectories = [self.working_directory_path / f"{problem_path.stem}.trajectory" for problem_path in fold_content["train"]]
                self.create_directories_content(fold_content["train"], fold_content["test"], train_set_trajectories)
                self.logger.debug("Finished creating fold!")
                yield self.train_set_dir_path, self.test_set_dir_path

                self.remove_created_directories()

            return

        problem_paths = []
        trajectory_paths = list(self.working_directory_path.glob(trajectory_suffix))
        trajectory_paths.sort()  # sort the trajectories so that the same order is used each time the algorithm runs
        items_per_fold = max_items if (0 < max_items <= len(trajectory_paths)) else len(trajectory_paths)
        trajectory_paths = random.sample(trajectory_paths, k=items_per_fold)
        for trajectory_file_path in trajectory_paths:
            problem_paths.append(self.working_directory_path / f"{trajectory_file_path.stem}.pddl")

        num_splits = len(trajectory_paths) if self.n_split == 0 else self.n_split
        for fold_index, (train_set_indices, test_set_indices) in enumerate(
            create_test_set_indices(len(problem_paths), num_splits, self.only_train_test)
        ):
            test_set_problems = [problem_paths[i] for i in test_set_indices]
            train_set_problems = [problem_paths[i] for i in train_set_indices]
            train_set_trajectories = [trajectory_paths[i] for i in range(len(trajectory_paths)) if i not in test_set_indices]
            self.logger.info(
                f"Created a new fold - train set has the following problems: {train_set_problems} "
                f"and test set had the following problems: {test_set_problems}"
            )
            self.create_directories_content(train_set_problems, test_set_problems, train_set_trajectories)
            folds_data[f"{FOLDS_LABEL}_{fold_index}"] = {
                "train": [str(p.absolute()) for p in train_set_problems],
                "test": [str(p.absolute()) for p in test_set_problems],
            }
            self.logger.debug("Finished creating fold!")
            yield self.train_set_dir_path, self.test_set_dir_path

            self.remove_created_directories()

        save_fold_settings(self.working_directory_path, folds_data)



"""Runs experiments for the numeric model learning algorithms."""
import argparse
import json
import os
import shutil
import signal
import subprocess
import time
import uuid
from pathlib import Path
from typing import List, Optional, Dict, Tuple, Any

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import Observation, Domain

from experiments.concurrent_execution.parallel_basic_experiment_runner import (
    ParallelExperimentRunner,
    configure_iteration_logger,
)
from experiments.experiments_consts import NUMERIC_SAM_ALGORITHM_VERSIONS
from sam_learning.core import LearnerDomain
from statistics import LearningStatisticsManager
from statistics.utils import init_semantic_performance_calculator
from trajectory_creators import PlanMinerTrajectoriesCreator
from utilities import LearningAlgorithmType
from validators import DomainValidator

PLAN_MINER_DIR_PATH = os.environ["PLAN_MINER_DIR_PATH"]
LEARNING_TIMEOUT = 600


class SingleIterationNSAMExperimentRunner(ParallelExperimentRunner):
    """Class to conduct offline numeric action model learning experiments."""

    def __init__(
        self,
        working_directory_path: Path,
        domain_file_name: str,
        polynom_degree: int,
        learning_algorithm: LearningAlgorithmType,
        fluents_map_path: Optional[Path],
        problem_prefix: str = "pfile",
        running_triplets_experiment: bool = False,
        plan_miner_workdir: Path = None,
        fold_num: int = 0,
    ):
        super().__init__(
            working_directory_path=working_directory_path,
            domain_file_name=domain_file_name,
            learning_algorithm=learning_algorithm,
            problem_prefix=problem_prefix,
            running_triplets_experiment=running_triplets_experiment,
        )
        self.fluents_map = None
        if fluents_map_path is not None:
            with open(fluents_map_path, "rt") as json_file:
                self.fluents_map = json.load(json_file)

        self.polynom_degree = polynom_degree
        self.fold_num = fold_num
        self.semantic_performance_calc = None
        self.plan_miner_domain_file_name = None
        self.plan_miner_workdir = plan_miner_workdir
        if learning_algorithm == LearningAlgorithmType.plan_miner:
            self.plan_miner_domain_file_name = f"{domain_file_name.split('.')[0]}_plan_miner.pddl"
            self.domain_validator = DomainValidator(
                working_directory_path=self.working_directory_path,
                learning_algorithm=learning_algorithm,
                reference_domain_path=self.plan_miner_workdir / self.plan_miner_domain_file_name,
                problem_prefix=problem_prefix,
            )

        else:
            self.domain_validator = DomainValidator(
                self.working_directory_path, learning_algorithm, self.working_directory_path / domain_file_name, problem_prefix=problem_prefix,
            )

    def _init_plan_miner_performance_evaluation(self) -> None:
        """Initializes the algorithm of the semantic precision - recall calculator - for plan miner only."""
        self.semantic_performance_calc = init_semantic_performance_calculator(
            working_directory_path=self.working_directory_path,
            domain_file_path=self.plan_miner_workdir / self.plan_miner_domain_file_name,
            learning_algorithm=self._learning_algorithm,
            test_set_dir_path=self.working_directory_path / "performance_evaluation_trajectories" / f"fold_{self.fold_num}",
            problem_prefix=self.problem_prefix,
            executing_agents=self.executing_agents,
        )

    def _handle_plan_miner_failure(self) -> Tuple[LearnerDomain, Dict[str, Any]]:
        plan_miner_domain = DomainParser(self.plan_miner_workdir / self.plan_miner_domain_file_name, partial_parsing=True).parse_domain()
        plan_miner_domain.actions = {}
        learned_domain = LearnerDomain(plan_miner_domain)
        return learned_domain, {"learning_time": -1}

    def _create_learned_domain_for_evaluation(self, plan_miner_output_domain_path: Path) -> LearnerDomain:
        """Creates the learned domain for evaluation purposes (The learner domain object).

        :param plan_miner_output_domain_path: the path to the Plan-Miner output domain.
        :return: the learned domain for evaluation.
        """
        self.logger.info("Creating the learned domain for PlanMiner's evaluation.")
        try:
            plan_miner_domain = DomainParser(plan_miner_output_domain_path).parse_domain()
            temp_domain = DomainParser(self.plan_miner_workdir / self.plan_miner_domain_file_name, partial_parsing=True).parse_domain()
            learned_domain = LearnerDomain(plan_miner_domain)
            learned_domain.name = temp_domain.name
            learned_domain.requirements = temp_domain.requirements

            for action_name, action_data in plan_miner_domain.actions.items():
                learned_domain.actions[action_name].signature = action_data.signature
                learned_domain.actions[action_name].preconditions = action_data.preconditions
                learned_domain.actions[action_name].discrete_effects = action_data.discrete_effects
                learned_domain.actions[action_name].numeric_effects = action_data.numeric_effects

        except Exception as e:
            self.logger.warning(f"Failed to parse the learned domain for PlanMiner. Probably PlanMiner created a domain with syntax errors.: {e}")
            plan_miner_domain = DomainParser(self.plan_miner_workdir / self.plan_miner_domain_file_name, partial_parsing=True).parse_domain()
            plan_miner_domain.actions = {}
            learned_domain = LearnerDomain(plan_miner_domain)

        with open(plan_miner_output_domain_path, "wt") as domain_file:
            # This should keep the domain name in the correct format and maintain the correct domain name in the PDDL file.
            domain_file.write(learned_domain.to_pddl(should_simplify=False))
            return learned_domain

    def _run_plan_miner_process(self, plan_miner_trajectory_file_path: Path) -> Optional[Path]:
        """Runs the Plan-Miner process to learn the action model.

        :param plan_miner_trajectory_file_path: the path to the Plan-Miner trajectory file.
        :return: the path to the learned action model (if successful).
        """
        plan_miner_domain_name = f"{self.domain_file_name.split('.')[0]}_{uuid.uuid4()}"
        os.chdir(PLAN_MINER_DIR_PATH)
        # cmd = ./PlanMiner {path to pts file} {domain name without extension}
        process = subprocess.Popen(f"./bin/PlanMiner {plan_miner_trajectory_file_path} {plan_miner_domain_name}", shell=True)
        try:
            process.wait(timeout=LEARNING_TIMEOUT)

        except subprocess.TimeoutExpired:
            self.logger.error("Plan-Miner failed to learn action model with the given timeout.")
            os.kill(process.pid, signal.SIGTERM)
            os.system("pkill -f ./bin/PlanMiner")

        except subprocess.CalledProcessError as e:
            self.logger.error(f"PlanMiner failed with the following return code: {e.returncode}")
            return None

        plan_miner_output_domain_path = Path(PLAN_MINER_DIR_PATH) / f"{plan_miner_domain_name}.pddl"
        if plan_miner_output_domain_path.exists():
            self.logger.info("PlanMiner finished successfully!")
            return plan_miner_output_domain_path

        self.logger.error("Plan-Miner failed to learn the action model.")
        return None

    def _apply_plan_miner(self, test_set_dir_path: Path) -> Tuple[LearnerDomain, Dict[str, Any]]:
        """Applies the learning algorithm Plan-Miner to learn the action model.

        :param test_set_dir_path: the path to the directory containing the test problems.
        :return: the learned action model and the learning statistics.
        """
        self.logger.info("Creating the formatted domain and the trajectories for Plan-Miner.")
        plan_miner_trajectory_creator = PlanMinerTrajectoriesCreator(
            domain_file_name=self.domain_file_name, working_directory_path=self.plan_miner_workdir
        )
        plan_miner_trajectory_creator.create_plan_miner_trajectories()
        # initializing the specific numeric performance calculator for Plan-Miner
        self._init_plan_miner_performance_evaluation()
        # Since only after this part the domain is created in the format that Plan-Miner can understand
        # we initialize the statistics manager here
        self.learning_statistics_manager = LearningStatisticsManager(
            working_directory_path=self.working_directory_path,
            domain_path=self.plan_miner_workdir / self.plan_miner_domain_file_name,
            learning_algorithm=self._learning_algorithm,
        )
        plan_miner_trajectory_file_path = self.plan_miner_workdir / "plan_miner_trajectories.pts"
        self.logger.info("Applying Plan-Miner to learn the action model.")
        plan_miner_start_time = time.time()
        plan_miner_output_domain_path = self._run_plan_miner_process(plan_miner_trajectory_file_path)
        learning_report = {
            "learning_time": time.time() - plan_miner_start_time,
        }
        if plan_miner_output_domain_path is None:
            self.logger.info("Plan-Miner failed to learn the action model.")
            return self._handle_plan_miner_failure()

        learned_domain = self._create_learned_domain_for_evaluation(plan_miner_output_domain_path)
        shutil.move(plan_miner_output_domain_path, test_set_dir_path / self.domain_file_name)
        return learned_domain, learning_report

    def _apply_learning_algorithm(
        self, partial_domain: Domain, allowed_observations: List[Observation], test_set_dir_path: Path
    ) -> Tuple[LearnerDomain, Dict[str, str]]:
        """Learns the action model using the numeric action model learning algorithms.

        :param partial_domain: the partial domain without the actions' preconditions and effects.
        :param allowed_observations: the allowed observations.
        :param test_set_dir_path: the path to the directory containing the test problems.
        :return: the learned action model and the learned action model's learning statistics.
        """
        if self._learning_algorithm == LearningAlgorithmType.plan_miner:
            return self._apply_plan_miner(test_set_dir_path)

        learner = NUMERIC_SAM_ALGORITHM_VERSIONS[self._learning_algorithm](
            partial_domain=partial_domain, polynomial_degree=self.polynom_degree, relevant_fluents=self.fluents_map
        )
        return learner.learn_action_model(allowed_observations)


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Runs the numeric action model learning algorithms evaluation experiments.")
    parser.add_argument("--working_directory_path", required=True, help="The path to the directory where the domain is")
    parser.add_argument("--domain_file_name", required=True, help="the domain file name including the extension")
    parser.add_argument(
        "--learning_algorithm", required=True, type=int, choices=[3, 5, 15, 19, 20, 21],
    )
    parser.add_argument(
        "--fluents_map_path", required=False, help="The path to the file mapping to the preconditions' " "fluents", default=None,
    )
    parser.add_argument(
        "--polynom_degree", required=False, help="The degree of the polynomial to set in the learning algorithm.", default=0,
    )
    parser.add_argument("--problems_prefix", required=False, help="The prefix of the problems' file names", type=str, default="pfile")
    parser.add_argument("--fold_number", required=True, help="The number of the fold to run", type=int)
    parser.add_argument("--iteration_number", required=True, help="The current iteration to execute", type=int)
    parser.add_argument("--debug", required=False, help="Whether in debug mode.", type=bool, default=False)
    args = parser.parse_args()
    return args


def main():
    args = parse_arguments()
    configure_iteration_logger(args)
    learning_algorithm = LearningAlgorithmType(args.learning_algorithm)
    working_directory_path = Path(args.working_directory_path)
    iteration_number = int(args.iteration_number)
    plan_miner_workdir = (working_directory_path / "train") / f"fold_{args.fold_number}_{args.learning_algorithm}_{iteration_number}"
    offline_learner = SingleIterationNSAMExperimentRunner(
        working_directory_path=working_directory_path,
        domain_file_name=args.domain_file_name,
        learning_algorithm=learning_algorithm,
        fluents_map_path=Path(args.fluents_map_path) if args.fluents_map_path else None,
        polynom_degree=int(args.polynom_degree),
        problem_prefix=args.problems_prefix,
        plan_miner_workdir=plan_miner_workdir,
        fold_num=int(args.fold_number),
    )
    offline_learner.run_fold_iteration(
        fold_num=args.fold_number,
        train_set_dir_path=(working_directory_path / "train") / f"fold_{args.fold_number}_{args.learning_algorithm}_{iteration_number}",
        test_set_dir_path=(working_directory_path / "test") / f"fold_{args.fold_number}_{args.learning_algorithm}_{iteration_number}",
        iteration_number=int(args.iteration_number),
    )


if __name__ == "__main__":
    main()



import sys
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def plot_results(results_directory_path: Path, file_template: str = "*solving_combined_statistics.csv", using_triplets: bool = False) -> None:
    """Plot the results of the experiments."""
    for file_path in results_directory_path.glob(file_template):
        df = pd.read_csv(file_path)

        # Define color-blind friendly palette
        color_palette = sns.color_palette("colorblind", n_colors=len(df["learning_algorithm"].unique()))
        line_styles = ["-", "--", "-.", ":", "--", "-.", ":"]
        marker_options = ["o", "s", "D", "v", "^", ">", "<", "p", "P", "*", "X", "d"]
        line_styles_iterator = list(line_styles)
        color_cycle = list(color_palette)
        markers = list(marker_options)

        # Group the data by 'num_trajectories', 'learning_algorithm' and calculate the mean and std of 'percent_ok'
        df = df[df["learning_algorithm"] != "incremental_nsam"]  # Remove max_percent_ok from the plot
        group_by_columns = ["num_trajectories", "learning_algorithm"] if not using_triplets else ["num_trajectory_triplets", "learning_algorithm"]
        grouped_data = df.groupby(group_by_columns, as_index=False).agg(avg_percent_ok=("percent_ok", "mean"), std_percent_ok=("percent_ok", "std"),)
        labels = {
            "numeric_sam": "NSAM*",
            "naive_nsam": "NSAM",
            "plan_miner": "Plan Miner",
        }

        # Plotting
        sns.set(style="whitegrid")
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(20, 8))

        legend_order = [0, 2, 1]

        x_lim = max(grouped_data["num_trajectories"]) * 1.25 if not using_triplets else max(grouped_data["num_trajectory_triplets"]) * 1.25
        # Plot a line for each learning algorithm
        for index, algo in enumerate(df["learning_algorithm"].unique()):
            algo_data = grouped_data[grouped_data["learning_algorithm"] == algo]
            plot_x_axis = algo_data["num_trajectories"] if not using_triplets else algo_data["num_trajectory_triplets"]

            ax1.plot(
                plot_x_axis,
                algo_data["avg_percent_ok"],
                linestyle=line_styles_iterator[index],
                label=labels[algo],
                marker=markers[index],
                color=color_cycle[index],
                linewidth=3,
            )

            # Plot standard deviation as shaded area around the mean line
            ax1.fill_between(
                plot_x_axis,
                np.clip(algo_data["avg_percent_ok"] - algo_data["std_percent_ok"], 0, 100),
                np.clip(algo_data["avg_percent_ok"] + algo_data["std_percent_ok"], 0, 100),
                alpha=0.2,
            )

            ax2.plot(
                plot_x_axis,
                algo_data["avg_percent_ok"],
                linestyle=line_styles_iterator[index],
                label=labels[algo],
                marker=markers[index],
                color=color_cycle[index],
                linewidth=3,
            )

            # Plot standard deviation as shaded area around the mean line
            ax2.fill_between(
                plot_x_axis,
                np.clip(algo_data["avg_percent_ok"] - algo_data["std_percent_ok"], 0, 100),
                np.clip(algo_data["avg_percent_ok"] + algo_data["std_percent_ok"], 0, 100),
                alpha=0.2,
            )

            # Set plot labels and title
            x_axis_label = "# Triplets" if using_triplets else "# Trajectories"
            ax1.set_xlabel(x_axis_label, fontsize=28)
            ax1.set_ylabel("AVG % of solved", fontsize=28)
            ax1.set_ylim(0, 100)
            ax1.tick_params(axis="both", which="major", labelsize=24)

            ax2.set_xlabel(f"{x_axis_label} (0-{int(x_lim * 0.1)})", fontsize=28)
            ax2.set_xlim(1, x_lim * 0.1)
            ax2.set_ylim(0, 100)
            ax2.tick_params(axis="both", which="major", labelsize=24)

        ax1.grid(True)
        ax2.grid(True)

        # Add a legend
        handles1, legend_labels1 = plt.gca().get_legend_handles_labels()
        ax1.legend([handles1[idx] for idx in legend_order], [legend_labels1[idx] for idx in legend_order], fontsize=28)

        handles2, legend_labels2 = plt.gca().get_legend_handles_labels()
        ax2.legend([handles2[idx] for idx in legend_order], [legend_labels2[idx] for idx in legend_order], fontsize=28)

        # Add a vertical dashed line at x = 5, with color changed to black
        ax1.axvline(x=x_lim * 0.1, color="black", linestyle="--", linewidth=4)
        # Add an annotation to show the zoom effect
        ax1.annotate(
            "Zoom up to\n here", xy=(x_lim * 0.1, 20), xytext=(x_lim * 0.15, 15), arrowprops=dict(facecolor="black", shrink=0.05), fontsize=18
        )

        output_file_path = file_path.parent / f"{file_path.stem}_plot.png"
        plt.savefig(output_file_path, bbox_inches="tight")


if __name__ == "__main__":
    results_path = Path(sys.argv[1])
    plot_results(results_path)



from .discrete_precision_recall_calculator import PrecisionRecallCalculator
from .learning_statistics_manager import LearningStatisticsManager
from .numeric_performance_calculator import NumericPerformanceCalculator
from .semantic_performance_calculator import SemanticPerformanceCalculator, _calculate_precision_recall
from .ma_performance_calculator import MASamPerformanceCalculator



import argparse
import csv
import logging
from pathlib import Path
from typing import List, Optional

from pddl_plus_parser.lisp_parsers import DomainParser
from pddl_plus_parser.models import Domain

from experiments.plotting.plot_nsam_results import plot_results
from experiments.plotting.plot_nsam_solo_results import plot_solo_results
from statistics.learning_statistics_manager import LEARNED_ACTIONS_STATS_COLUMNS
from statistics.numeric_performance_calculator import NUMERIC_PERFORMANCE_STATS
from statistics.semantic_performance_calculator import SEMANTIC_PRECISION_STATS
from utilities import LearningAlgorithmType
from validators.safe_domain_validator import SOLVING_STATISTICS

FOLD_FIELD = "fold"


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates the folds directories for the slurm tasks to work on.")
    parser.add_argument("--working_directory_path", required=True, help="The path to the directory where the domain is")
    parser.add_argument("--domain_file_name", required=True, help="the domain file name including the extension")
    parser.add_argument("--learning_algorithms", required=True, help="the list of algorithms that will run in parallel")
    parser.add_argument("--num_folds", required=True, help="the number of folds to that were created", type=int, default=5)
    parser.add_argument("--internal_iterations", required=True, help="The internal iterations that the algorithm will run in parallel.")
    args = parser.parse_args()
    return args


class DistributedResultsCollector:
    logger: logging.Logger

    def __init__(
        self,
        working_directory_path: Path,
        domain_file_name: str,
        learning_algorithms: List[str] = None,
        num_folds: int = 5,
        iterations: List[int] = None,
    ):
        self.domain_file_name = domain_file_name
        self.working_directory_path = working_directory_path
        self.learning_algorithms = [LearningAlgorithmType(int(e)) for e in learning_algorithms]
        self.num_folds = num_folds
        self.iterations = iterations
        self.logger = logging.getLogger("ClusterRunner")

    def _combine_statistics_data(
        self,
        file_path_template: str,
        combined_statistics_data: List[dict],
        exclude_algorithm: Optional[LearningAlgorithmType] = None,
        using_triplets: bool = False,
    ) -> None:
        """Combines the statistics data from the statistics files in the results directory to a single file.

        :param file_path_template: the template of the file path to use to find the statistics files.
        :param combined_statistics_data: the list to append the combined statistics data to.
        :param exclude_algorithm: the algorithm to exclude and not include in the combined statistics.
        :param using_triplets: whether to use triplets or not.
        """
        results_directory = self.working_directory_path / "results_directory"
        algorithms_to_iterate = (
            self.learning_algorithms
            if exclude_algorithm is None
            else [algorithm for algorithm in self.learning_algorithms if algorithm != exclude_algorithm]
        )
        for fold in range(self.num_folds):
            for iteration in self.iterations:
                for learning_algorithm in algorithms_to_iterate:
                    solving_statistics_file_path = (
                        results_directory / file_path_template.format(fold=fold, iteration=iteration, learning_algorithm=learning_algorithm.name)
                        if not using_triplets
                        else results_directory / file_path_template.format(fold=fold, learning_algorithm=learning_algorithm.name)
                    )
                    with open(solving_statistics_file_path, "rt") as statistics_file:
                        reader = csv.DictReader(statistics_file)
                        combined_statistics_data.extend([{FOLD_FIELD: fold, **row} for row in reader])

    def _collect_solving_statistics(self, collecting_triplets: bool = False) -> None:
        """Collects the statistics from the statistics files in the results directory based on the given parameters.

        :param collecting_triplets: whether to collect the statistics based on triplets or not.
        """
        self.logger.info("Collecting the solving statistics from the results directory.")
        results_directory = self.working_directory_path / "results_directory"
        combined_statistics_file_path = (
            results_directory / "solving_combined_statistics.csv"
            if not collecting_triplets
            else results_directory / "solving_combined_statistics_with_triplets.csv"
        )
        combined_statistics_data = []
        file_path_template = (
            "{learning_algorithm}_problem_solving_stats_fold_{fold}_{iteration}_trajectories.csv"
            if not collecting_triplets
            else "{learning_algorithm}_problem_solving_stats_fold_{fold}.csv"
        )
        self._combine_statistics_data(file_path_template, combined_statistics_data, using_triplets=collecting_triplets)

        with open(combined_statistics_file_path, "wt") as combined_statistics_file:
            writer = csv.DictWriter(combined_statistics_file, fieldnames=[FOLD_FIELD, *SOLVING_STATISTICS])
            writer.writeheader()
            writer.writerows(combined_statistics_data)

    def _collect_numeric_semantic_performance_statistics(self, collecting_triplets: bool = False) -> None:
        self.logger.info("Collecting the numeric performance statistics from the results directory.")
        domain = DomainParser(self.working_directory_path / self.domain_file_name).parse_domain()
        combined_statistics_file_path = (
            self.working_directory_path / "results_directory" / "numeric_performance_combined_statistics.csv"
            if not collecting_triplets
            else self.working_directory_path / "results_directory" / "numeric_performance_combined_statistics_with_triplets.csv"
        )
        file_path_template = (
            "{learning_algorithm}_" + domain.name + "_numeric_learning_performance_stats_fold_{fold}_{iteration}.csv"
            if not collecting_triplets
            else "{learning_algorithm}_" + domain.name + "_numeric_learning_performance_stats_fold_{fold}.csv"
        )
        combined_statistics_data = []
        self._combine_statistics_data(file_path_template, combined_statistics_data, using_triplets=collecting_triplets)
        with open(combined_statistics_file_path, "wt") as combined_statistics_file:
            writer = csv.DictWriter(combined_statistics_file, fieldnames=[FOLD_FIELD, *NUMERIC_PERFORMANCE_STATS])
            writer.writeheader()
            writer.writerows(combined_statistics_data)
        self.logger.info("Done collecting the statistics from the results directory!")

    def _collect_semantic_performance_statistics(
        self, domain: Domain, exclude_algorithm: Optional[LearningAlgorithmType] = None, using_triplets: bool = False
    ) -> None:
        """Collects the semantic performance statistics from the results directory.

        :param domain: the domain to collect the statistics for.
        """
        self.logger.info("Collecting the semantic performance statistics from the results directory.")
        combined_semantic_performance_file_path = (
            self.working_directory_path / "results_directory" / "semantic_performance_combined_statistics.csv"
            if not using_triplets
            else self.working_directory_path / "results_directory" / "semantic_performance_combined_statistics_with_triplets.csv"
        )
        combined_semantic_performance_statistics_data = []
        file_path_template = (
            "{learning_algorithm}_" + domain.name + "_{fold}_{iteration}_semantic_performance.csv"
            if not using_triplets
            else "{learning_algorithm}_" + domain.name + "_{fold}_semantic_performance.csv"
        )
        self._combine_statistics_data(file_path_template, combined_semantic_performance_statistics_data, exclude_algorithm=exclude_algorithm)
        with open(combined_semantic_performance_file_path, "wt") as combined_statistics_file:
            writer = csv.DictWriter(combined_statistics_file, fieldnames=[FOLD_FIELD, *SEMANTIC_PRECISION_STATS])
            writer.writeheader()
            writer.writerows(combined_semantic_performance_statistics_data)

    def _collect_syntactic_performance_statistics(
        self, domain: Domain, exclude_algorithm: Optional[LearningAlgorithmType] = None, using_triplets: bool = False
    ) -> None:
        """Collects the syntactic performance statistics from the results directory.

        :param domain: the domain to collect the statistics for.
        """
        self.logger.info("Collecting the syntactic performance statistics from the results directory.")
        combined_syntactic_performance_file_path = (
            self.working_directory_path / "results_directory" / "syntactic_performance_combined_statistics.csv"
            if not using_triplets
            else self.working_directory_path / "results_directory" / "syntactic_performance_combined_statistics_with_triplets.csv"
        )
        combined_action_performance_statistics_data = []
        action_performance_path_template = (
            "{learning_algorithm}_" + domain.name + "_action_stats_fold_{fold}_{iteration}.csv"
            if not using_triplets
            else "{learning_algorithm}_" + domain.name + "_action_stats_fold_{fold}.csv"
        )
        self._combine_statistics_data(
            action_performance_path_template, combined_action_performance_statistics_data, exclude_algorithm=exclude_algorithm
        )
        with open(combined_syntactic_performance_file_path, "wt") as combined_statistics_file:
            writer = csv.DictWriter(combined_statistics_file, fieldnames=[FOLD_FIELD, *LEARNED_ACTIONS_STATS_COLUMNS])
            writer.writeheader()
            writer.writerows(combined_action_performance_statistics_data)
